"""SQLAlchemy simple column generation."""

import typing

import sqlalchemy

from ... import exceptions
from ... import helpers
from ... import types as oa_types
from . import types


def construct(*, artifacts: oa_types.SimplePropertyArtifacts) -> types.Column:
    """
    Construct column from artifacts.

    Args:
        artifacts: The artifacts of the column.

    Returns:
        The SQLAlchemy column.

    """
    type_ = _determine_type(artifacts=artifacts)
    foreign_key: typing.Optional[types.ForeignKey] = None
    if artifacts.extension.foreign_key is not None:
        foreign_key_kwargs: oa_types.TKwargs = {}
        if artifacts.extension.foreign_key_kwargs is not None:
            foreign_key_kwargs = artifacts.extension.foreign_key_kwargs
        foreign_key = types.ForeignKey(
            artifacts.extension.foreign_key, **foreign_key_kwargs
        )
    # Map default value
    default = None
    if artifacts.open_api.default is not None:
        default = helpers.oa_to_py_type.convert(
            value=artifacts.open_api.default,
            type_=artifacts.open_api.type,
            format_=artifacts.open_api.format,
        )

    # Calculate server default
    server_default = None
    if artifacts.extension.server_default is not None:
        server_default = sqlalchemy.text(artifacts.extension.server_default)

    # Calculate nullable
    nullable = helpers.calculate_nullable(
        nullable=artifacts.open_api.nullable,
        generated=artifacts.extension.autoincrement is True,
        defaulted=default is not None or artifacts.extension.server_default is not None,
        required=artifacts.required,
    )

    # Generate optional keyword arguments
    opt_kwargs: types.TOptColumnArgs = {}
    if artifacts.extension.primary_key is not None:
        opt_kwargs["primary_key"] = artifacts.extension.primary_key
    if artifacts.extension.autoincrement is not None:
        opt_kwargs["autoincrement"] = artifacts.extension.autoincrement
    if artifacts.extension.index is not None:
        opt_kwargs["index"] = artifacts.extension.index
    if artifacts.extension.unique is not None:
        opt_kwargs["unique"] = artifacts.extension.unique
    # Generate kwargs
    kwargs: oa_types.TKwargs = {}
    if artifacts.extension.kwargs is not None:
        kwargs = artifacts.extension.kwargs
    return types.Column(
        type_,
        foreign_key,
        nullable=nullable,
        default=default,
        server_default=server_default,
        **opt_kwargs,
        **kwargs,
    )


def _determine_type(*, artifacts: oa_types.SimplePropertyArtifacts) -> types.Type:
    """
    Determine the type for a specification.

    Raise FeatureNotImplementedError for unsupported types.

    Args:
        artifacts: The artifacts for the column.

    Returns:
        The type for the column.

    """
    # Determining the type
    if artifacts.open_api.type == "integer":
        return _handle_integer(artifacts=artifacts)
    if artifacts.open_api.type == "number":
        return _handle_number(artifacts=artifacts)
    if artifacts.open_api.type == "string":
        return _handle_string(artifacts=artifacts)
    if artifacts.open_api.type == "boolean":
        return _handle_boolean()

    raise exceptions.FeatureNotImplementedError(
        f"{artifacts.open_api.type} has not been implemented"
    )


def _handle_integer(
    *, artifacts: oa_types.SimplePropertyArtifacts
) -> typing.Union[types.Integer, types.BigInteger]:
    """
    Handle artifacts for an integer type.

    Raise FeatureNotImplementedError if a format that is not supported is defined.

    Args:
        artifacts: The artifacts for the column.

    Returns:
        The SQLAlchemy integer type of the column.

    """
    if artifacts.open_api.format is None or artifacts.open_api.format == "int32":
        return types.Integer()
    if artifacts.open_api.format == "int64":
        return types.BigInteger()
    raise exceptions.FeatureNotImplementedError(
        f"{artifacts.open_api.format} format for integer is not supported."
    )


def _handle_number(*, artifacts: oa_types.SimplePropertyArtifacts) -> types.Number:
    """
    Handle artifacts for an number type.

    Raise FeatureNotImplementedError if a format that is not supported is defined.

    Args:
        artifacts: The artifacts for the column.

    Returns:
        The SQLAlchemy number type of the column.

    """
    if artifacts.open_api.format is None or artifacts.open_api.format == "float":
        return types.Number()
    raise exceptions.FeatureNotImplementedError(
        f"{artifacts.open_api.format} format for number is not supported."
    )


def _handle_string(
    *, artifacts: oa_types.SimplePropertyArtifacts
) -> typing.Union[types.String, types.Binary, types.Date, types.DateTime]:
    """
    Handle artifacts for an string type.

    Raise FeatureNotImplementedError if a format that is not supported is defined.

    Args:
        artifacts: The artifacts for the column.

    Returns:
        The SQLAlchemy string type of the column.

    """
    if artifacts.open_api.format == "binary":
        if artifacts.open_api.max_length is None:
            return types.Binary()
        return types.Binary(length=artifacts.open_api.max_length)
    if artifacts.open_api.format == "date":
        return types.Date()
    if artifacts.open_api.format == "date-time":
        return types.DateTime()
    if artifacts.open_api.max_length is None:
        return types.String()
    return types.String(length=artifacts.open_api.max_length)


def _handle_boolean() -> types.Boolean:
    """
    Handle artifacts for an boolean type.

    Args:
        artifacts: The artifacts for the column.

    Returns:
        The SQLAlchemy boolean type of the column.

    """
    return types.Boolean()



"""Check unmanaged schemas for why they are unmanaged."""

import typing

from ... import exceptions
from ... import helpers as _oa_helpers
from ... import types as _oa_types
from .. import helpers as _helpers
from . import spec_validation
from . import types


def _reason_to_result(*, reason: str) -> types.TResult:
    """Convert a reason to a result."""
    return {"valid": False, "reason": reason}


def check_model(schemas: _oa_types.Schemas, schema: _oa_types.Schema) -> types.TModel:
    """
    Check a model.

    Algorithm:
    1. check for type
    2. check that type is object and
    3. check for tablename.

    Args:
        schema: The schema of the model to check.

    Returns:
        Whether the model and its properties are valid with a reason if it is not.

    """
    try:
        # Check type
        type_ = _oa_helpers.peek.peek_key(schema=schema, schemas=schemas, key="type")
        if type_ is None:
            return {
                "result": _reason_to_result(
                    reason='no "type" key was found, define a type'
                )
            }
        if not isinstance(type_, str):
            return {
                "result": _reason_to_result(
                    reason="the type value is True, change it to a string value"
                )
            }
        if type_ != "object":
            return {
                "result": _reason_to_result(
                    reason=(
                        f'the type of the schema is "{type_}", change it to be "object"'
                    )
                )
            }

    except exceptions.MalformedSchemaError as exc:
        return {"result": _reason_to_result(reason=f"malformed schema :: {exc}")}
    except exceptions.SchemaNotFoundError as exc:
        return {"result": _reason_to_result(reason=f"reference :: {exc}")}

    # tablename is not defined
    return {
        "result": _reason_to_result(
            reason='no "x-tablename" key was found, define the name of the table'
        )
    }


def check_models(*, schemas: _oa_types.Schemas) -> types.TModels:
    """
    Check the models of a schema.

    Assume the schemas is valid although any of its models may not.

    Args:
        schemas: The schemas to check.

    Returns:
        The result for each model.

    """
    not_constructables = _helpers.iterate.not_constructable(schemas=schemas)
    not_constructables_result = map(
        lambda args: (args[0], check_model(schemas, args[1])), not_constructables
    )
    return dict(not_constructables_result)


def check(*, spec: typing.Any) -> types.TSpec:
    """
    Check a specification.

    Args:
        spec: The specification to check.

    Returns:
        Whether the specification is valid with a reason if it is not.

    """
    # Check spec to schemas
    spec_result = spec_validation.check(spec=spec)
    if not spec_result.valid:
        return {"result": types.t_result_from_result(spec_result)}

    # Check that there is at least 1 model
    assert isinstance(spec, dict)
    components = spec.get("components")
    assert isinstance(components, dict)
    schemas = components.get("schemas")
    assert isinstance(schemas, dict)

    return {"result": {"valid": True}, "models": check_models(schemas=schemas)}



"""Types for UtilityBase."""

import datetime
import typing

from .. import types as oa_types

# Types for converting to dictionary
TSimpleDict = typing.Union[int, float, str, bool]
TOptSimpleDict = typing.Optional[TSimpleDict]
TObjectDict = typing.Dict[str, typing.Any]
TOptObjectDict = typing.Optional[TObjectDict]
TArrayDict = typing.List[TOptObjectDict]
TOptArrayDict = typing.Optional[TArrayDict]
TComplexDict = typing.Union[TOptObjectDict, TOptArrayDict]
TAnyDict = typing.Union[TComplexDict, TOptSimpleDict]
# Types for converting from a dictionary
TStringCol = typing.Union[str, bytes, datetime.date, datetime.datetime]
TSimpleCol = typing.Union[int, float, TStringCol, bool]
TOptSimpleCol = typing.Optional[TSimpleCol]
TObjectCol = typing.Any  # pylint: disable=invalid-name
TOptObjectCol = typing.Optional[TObjectCol]
TArrayCol = typing.Iterable[TOptObjectCol]
TOptArrayCol = typing.Optional[TArrayCol]
TComplexCol = typing.Union[TOptObjectCol, TOptArrayCol]
TAnyCol = typing.Union[TComplexCol, TSimpleCol]


class TModel(oa_types.Protocol):
    """Defines interface for a model."""

    def to_dict(self) -> TObjectDict:
        """Interface for to_dict."""
        ...



"""Merges objects under allOf statement."""

import typing

from open_alchemy import types

from . import ref


def merge(
    *,
    schema: types.Schema,
    schemas: types.Schemas,
    skip_name: typing.Optional[str] = None,
) -> types.Schema:
    """
    Merge schemas under allOf statement.

    Merges schemas under allOf statement which is expected to have a list of schemas.
    Any duplicate keys will be overridden. Schemas are processed in the order they are
    listed.

    Args:
        schema: The schema to operate on.
        schemas: Used to resolve any $ref.
        skip_name (optional): The name of any schema to skip.

    Returns:
        The schema with all top level allOf statements resolved.

    """
    return _merge(schema, schemas, skip_name)


def _merge(
    schema: types.Schema, schemas: types.Schemas, skip_name: typing.Optional[str]
) -> types.Schema:
    """Implement merge."""
    all_of = schema.get("allOf")
    if all_of is None:
        return schema

    merged_schema: types.Schema = {}
    for sub_schema in all_of:
        # Resolving any $ref
        _, ref_schema = ref.resolve(
            name="", schema=sub_schema, schemas=schemas, skip_name=skip_name
        )
        # Merging any nested allOf
        merged_sub_schema = _merge(ref_schema, schemas, skip_name)

        # Capturing required arrays
        merged_required = merged_schema.get("required")
        sub_required = merged_sub_schema.get("required")
        # Capturing properties
        merged_properties = merged_schema.get("properties")
        sub_properties = merged_sub_schema.get("properties")
        # Capturing backrefs
        merged_backrefs = merged_schema.get("x-backrefs")
        sub_backrefs = merged_sub_schema.get("x-backrefs")

        # Combining sub into merged specification
        merged_schema = {**merged_schema, **merged_sub_schema}

        # Checking whether required was present on both specs
        if merged_required is not None and sub_required is not None:
            # Both have a required array, need to merge them together
            required_set = set(merged_required).union(sub_required)
            merged_schema["required"] = list(required_set)

        # Checking whether properties was present on both specs
        if merged_properties is not None and sub_properties is not None:
            # Both have properties, merge properties
            merged_schema["properties"] = {**merged_properties, **sub_properties}

        # Checking whether backrefs was present on both specs
        if merged_backrefs is not None and sub_backrefs is not None:
            # Both have backrefs, merge backrefs
            merged_schema["x-backrefs"] = {**merged_backrefs, **sub_backrefs}

    return merged_schema



"""Pre-process schemas by adding any association tables as a schema to schemas."""

import typing

from .. import helpers as oa_helpers
from .. import types
from . import helpers


def _assert_is_string(value: typing.Any) -> str:
    """Assert that a value is a string."""
    assert isinstance(value, str)
    return value


def _get_association_tablenames(
    *,
    association_schemas: typing.List[types.TNameSchema],
) -> typing.Set[str]:
    """
    Get the tablenames of the associations.

    Args:
        association_schemas: All the association schemas.

    Returns:
        A set of tablenames from the associations.

    """
    tablenames = map(
        lambda association: oa_helpers.peek.tablename(
            schema=association.schema, schemas={}
        ),
        association_schemas,
    )
    str_tablenames = map(_assert_is_string, tablenames)
    return set(str_tablenames)


class _TParentAllNames(typing.NamedTuple):
    """The name of the parent schema and all schemas with the same tablename."""

    parent_name: str
    all_names: typing.List[str]


_TTablenameParentAllNames = typing.Dict[str, _TParentAllNames]


def _get_tablename_schema_names(
    *, schemas: types.Schemas, tablenames: typing.Set[str]
) -> _TTablenameParentAllNames:
    """
    Get a mapping of tablenames to all schema names with that tablename.

    The algorithm is:
    1. Get a mapping of the tablename to the parent schema name by skipping any single
        table inheritance children.
    2. Get a list of schema names and tablenames and filter for those that are in the
        mapping.
    3. Combine any schema names with the same tablename and lookup the parent schema
        name.

    Args:
        schemas: All defines schemas.
        tablenames: All tablenames to filter for.

    Returns:
        A mapping of association tablenames to all schema names using that tablename
        and the parent schema name of a tablename.

    """
    # Get mapping of tablename to parent schema name
    constructables = helpers.iterate.constructable(schemas=schemas)
    not_single_inheritance_constructables = filter(
        lambda args: oa_helpers.inheritance.calculate_type(
            schema=args[1], schemas=schemas
        )
        != oa_helpers.inheritance.Type.SINGLE_TABLE,
        constructables,
    )
    tablename_parent_name_map = dict(
        map(
            lambda args: (
                oa_helpers.peek.prefer_local(
                    get_value=oa_helpers.peek.tablename, schema=args[1], schemas=schemas
                ),
                args[0],
            ),
            not_single_inheritance_constructables,
        )
    )

    # Get a list of schema names and tablenames which appear in the mapping
    constructables = helpers.iterate.constructable(schemas=schemas)
    name_tablenames = map(
        lambda args: (
            args[0],
            oa_helpers.peek.prefer_local(
                get_value=oa_helpers.peek.tablename, schema=args[1], schemas=schemas
            ),
        ),
        constructables,
    )
    filtered_name_tablenames = filter(
        lambda args: args[1] in tablenames, name_tablenames
    )

    mapping: _TTablenameParentAllNames = {}
    for name, tablename in filtered_name_tablenames:
        if tablename not in mapping:
            mapping[tablename] = _TParentAllNames(
                parent_name=tablename_parent_name_map[tablename], all_names=[]
            )
        mapping[tablename].all_names.append(name)

    return mapping


class _TParentNameForeignKeys(typing.NamedTuple):
    """The name of the parent schema and all foreign keys on the same tablename."""

    parent_name: str
    foreign_keys: typing.Set[str]


_TTablenameForeignKeys = typing.Dict[str, _TParentNameForeignKeys]


def _get_tablename_foreign_keys(
    tablename_parent_all_names: _TTablenameParentAllNames, schemas: types.Schemas
) -> _TTablenameForeignKeys:
    """
    Get a mapping of tablename to foreign keys defined on that tablename.

    Algorithm:
    1. Combine all the schema names on the same tablename using allOf and get all the
        properties on that tablename.
    2. Retrieve the value of x-foreign-key on each property and filter for those that
        are strings.
    3. Combine into a mapping of tablename against the foreign keys and parent name.

    Args:
        tablename_parent_all_names: Mapping of tablename to schema names.
        schemas: All defined schemas.

    Returns:
        Mapping of tablename to the parent schema and all foreign keys defined on that
        tablename.

    """
    # Get all the properties on a tablename
    tablename_schemas = map(
        lambda args: (
            args[0],
            {
                "allOf": [
                    {"$ref": f"#/components/schemas/{name}"}
                    for name in args[1].all_names
                ]
            },
        ),
        tablename_parent_all_names.items(),
    )
    tablename_properties = map(
        lambda args: (
            args[0],
            helpers.iterate.properties_items(schema=args[1], schemas=schemas),
        ),
        tablename_schemas,
    )

    # Get all the foreign keys of the properties on the tablename
    tablename_foreign_keys = map(
        lambda args: (
            args[0],
            map(
                lambda property_: oa_helpers.peek.prefer_local(
                    get_value=oa_helpers.peek.foreign_key,
                    schema=property_[1],
                    schemas=schemas,
                ),
                args[1],
            ),
        ),
        tablename_properties,
    )
    tablename_str_foreign_keys = map(
        lambda args: (
            args[0],
            filter(lambda foreign_key: isinstance(foreign_key, str), args[1]),
        ),
        tablename_foreign_keys,
    )
    tablename_foreign_key_set = map(
        lambda args: (args[0], set(args[1])), tablename_str_foreign_keys
    )

    # Combine the tablename against the parent name and foreign keys
    return dict(
        map(
            lambda args: (
                args[0],
                _TParentNameForeignKeys(
                    parent_name=tablename_parent_all_names[args[0]].parent_name,
                    foreign_keys=args[1],
                ),
            ),
            tablename_foreign_key_set,
        )
    )


def _combine_defined_expected_schema(
    parent_name_foreign_keys: _TParentNameForeignKeys,
    expected_schema: types.TNameSchema,
    schemas: types.Schemas,
) -> types.TNameSchema:
    """
    Combine defined and expected association schema.

    Algorithm:
    1. For the expected schema:
        a. Remove required array
        b. Filter the properties where their foreign is already defined
    2. Combine the expected schema in an allOf with the defined schema parent name.
    3. Change the name to the parent name.

    Args:
        parent_name_foreign_keys: The parent name and defined foreign keys of the
            defined association schema.
        expected_schema: The expected schema.

    Returns:
        The combined expected and defined schema.

    """
    expected_schema_value = expected_schema.schema
    del expected_schema_value["required"]
    expected_schema_value["properties"] = {
        property_[0]: property_[1]
        for property_ in expected_schema_value["properties"].items()
        if oa_helpers.peek.foreign_key(schema=property_[1], schemas={})
        not in parent_name_foreign_keys.foreign_keys
    }
    return types.TNameSchema(
        name=parent_name_foreign_keys.parent_name,
        schema={
            "allOf": [
                expected_schema_value,
                schemas[parent_name_foreign_keys.parent_name],
            ]
        },
    )


def _combine_defined_expected_schemas(
    *, association_schemas: typing.List[types.TNameSchema], schemas: types.Schemas
) -> typing.Iterator[types.TNameSchema]:
    """
    Combine all association schemas with any defined schemas.

    Args:
        association_schemas: All expected association schemas.
        schemas: All defined schemas.

    Returns:
        The association schemas merged with any existing schemas.

    """
    association_tablenames = _get_association_tablenames(
        association_schemas=association_schemas
    )
    tablename_schema_names = _get_tablename_schema_names(
        schemas=schemas, tablenames=association_tablenames
    )
    tablename_foreign_keys = _get_tablename_foreign_keys(
        tablename_parent_all_names=tablename_schema_names, schemas=schemas
    )

    for association_schema in association_schemas:
        association_tablename = oa_helpers.peek.tablename(
            schema=association_schema.schema, schemas={}
        )
        if association_tablename not in tablename_foreign_keys:
            yield association_schema
            continue

        yield _combine_defined_expected_schema(
            tablename_foreign_keys[association_tablename], association_schema, schemas
        )


def process(*, schemas: types.Schemas) -> None:
    """
    Pre-process the schemas to add association schemas as necessary.

    Algorithm:
    1. Iterate over all schemas and their properties retaining the parent schema and the
        schemas in the context and staying within the model properties only
    2. Filter for properties that (1) are relationships and (2) are many-to-many
        relationships
    3. Convert the property schema to an association schema
    4. Add them to the schemas

    Args:
        schemas: The schemas to process.

    """
    association_properties = helpers.association.get_association_property_iterator(
        schemas=schemas
    )
    association_schemas = list(
        map(
            lambda args: helpers.association.calculate_schema(
                property_schema=args.property.schema,
                parent_schema=args.parent.schema,
                schemas=schemas,
            ),
            association_properties,
        )
    )
    combined_association_schemas = _combine_defined_expected_schemas(
        association_schemas=association_schemas, schemas=schemas
    )
    for association in combined_association_schemas:
        schemas[association.name] = association.schema



"""SQLAlchemy json column generation."""

import typing

from ... import helpers
from ... import types as oa_types
from . import types


def construct(*, artifacts: oa_types.JsonPropertyArtifacts) -> types.Column:
    """
    Construct column from artifacts.

    Args:
        artifacts: The artifacts of the column.

    Returns:
        The SQLAlchemy column.

    """
    type_ = types.JSON()
    foreign_key: typing.Optional[types.ForeignKey] = None
    if artifacts.extension.foreign_key is not None:
        foreign_key_kwargs: oa_types.TKwargs = {}
        if artifacts.extension.foreign_key_kwargs is not None:
            foreign_key_kwargs = artifacts.extension.foreign_key_kwargs
        foreign_key = types.ForeignKey(
            artifacts.extension.foreign_key, **foreign_key_kwargs
        )

    # Calculate nullable
    nullable = helpers.calculate_nullable(
        nullable=artifacts.open_api.nullable,
        generated=False,
        defaulted=False,
        required=artifacts.required,
    )

    # Generate optional keyword arguments
    opt_kwargs: types.TOptColumnArgs = {}
    if artifacts.extension.primary_key is not None:
        opt_kwargs["primary_key"] = artifacts.extension.primary_key
    if artifacts.extension.index is not None:
        opt_kwargs["index"] = artifacts.extension.index
    if artifacts.extension.unique is not None:
        opt_kwargs["unique"] = artifacts.extension.unique
    # Generate kwargs
    kwargs: oa_types.TKwargs = {}
    if artifacts.extension.kwargs is not None:
        kwargs = artifacts.extension.kwargs
    return types.Column(
        type_,
        foreign_key,
        nullable=nullable,
        **opt_kwargs,
        **kwargs,
    )



"""Generate a model class based on artifacts."""

import typing

from open_alchemy import schemas
from open_alchemy import types as oa_types

from .. import artifacts as models_file_artifacts
from . import source as _source


def generate(*, artifacts: schemas.artifacts.types.ModelArtifacts, name: str) -> str:
    """
    Generate the class source from the schema.

    Args:
        schema: The schema of the model.
        name: The name of the model.

    Returns:
        The source code for the model class.

    """
    model_artifacts = models_file_artifacts.calculate(artifacts=artifacts, name=name)
    return _source.generate(artifacts=model_artifacts)



"""Gather artifacts for a property."""

from .... import helpers as oa_helpers
from .... import types as oa_types
from .. import types
from . import backref
from . import json
from . import relationship
from . import simple


def get(
    schemas: oa_types.Schemas,
    model_schema: oa_types.Schema,
    property_name: str,
    schema: oa_types.Schema,
    required: bool,
) -> types.TAnyPropertyArtifacts:
    """
    Retrieve the artifacts for a property.

    Args:
        schemas: All the defined schemas.
        model_schema: The schema that contains the property.
        property_name: The name of the property.
        schema: The schema of the property to gather artifacts for.
        required: WHether the property appears in the required list.

    Returns:
        The artifacts for the property.

    """
    type_ = oa_helpers.property_.calculate_type(schema=schema, schemas=schemas)

    if type_ == oa_types.PropertyType.SIMPLE:
        return simple.get(schemas, schema, required)

    if type_ == oa_types.PropertyType.JSON:
        return json.get(schemas, schema, required)

    if type_ == oa_types.PropertyType.BACKREF:
        return backref.get(schemas, schema)

    assert type_ == oa_types.PropertyType.RELATIONSHIP
    return relationship.get(schemas, model_schema, property_name, schema, required)



"""Convert object dictionary to column value."""

from ... import exceptions
from ... import facades
from ... import helpers
from ... import types as oa_types
from .. import types


def convert(
    value: types.TObjectDict, *, schema: oa_types.Schema
) -> types.TOptObjectCol:
    """
    Convert dictionary value to model instance.

    Raises InvalidInstanceError of the value is not a dictionary.

    Args:
        value: The value to convert.
        schema: The schema for the value.

    Returns:
        The converted value.

    """
    ref_model_name = helpers.ext_prop.get(source=schema, name="x-de-$ref")
    if ref_model_name is None:
        raise exceptions.MalformedSchemaError(
            "To construct object parameters the schema for the property must "
            "include the x-de-$ref extension property with the name of the "
            "model to construct for the property."
        )
    if not isinstance(value, dict):
        raise exceptions.InvalidInstanceError(
            "The value for an object parameter must be a dictionary."
        )
    ref_model = facades.models.get_model(name=ref_model_name)
    if ref_model is None:
        raise exceptions.SchemaNotFoundError(
            f"The referenced model {ref_model} was not found in the models."
        )
    return ref_model.from_dict(**value)



"""Helpers to support inheritance."""

import enum
import typing

from .. import exceptions
from .. import facades
from .. import types
from . import all_of as all_of_helper
from . import ext_prop as ext_prop_helper
from . import peek as peek_helper
from . import ref as ref_helper
from . import schema as schema_helper


def check_parent(
    *, schema: types.Schema, parent_name: str, schemas: types.Schemas
) -> bool:
    """
    Check that the parent is in the inheritance chain of a schema.

    Recursive function. The base cases are:
    1. the schema has $ref where the name matches the parent name if
        a. the referenced schema is constructable, return True or
        b. if the referenced schema is not constructable, return False,
    2. the schema does not have $ref nor allOf in which case return False and
    3. The schema has $ref but the referenced schema does not inherit.

    The recursive cases are:
    1. the schema has $ref where the name does not match the parent name in which case
        the function is called with the referenced schema and
    2. the schema has allOf where the function is called on each element in allOf and,
        if True is returned for any of them, True is returned otherwise False is
        returned.

    Raise MalformedSchemaError is the value of $ref is not a string.
    Raise MalformedSchemaError if the value of allOf is not a list.
    Raise MalformedSchemaError if the parent is not found in the chain.
    Raise MalformedSchemaError if the parent does not have x-tablename nor x-inherits.
    Raise MalformedSchemaError if a $ref value is seen again.

    Args:
        schema: The schema to check.
        parent_name: The parent to check for in the inheritance chain.
        schemas: All the schemas.
        seen_refs: The $ref that have already been check to avoid circular dependencies.

    Returns:
        Whether the parent is in the inheritance chain.

    """
    return _check_parent(schema, parent_name, schemas, set())


def _check_parent(
    schema: types.Schema,
    parent_name: str,
    schemas: types.Schemas,
    seen_refs: typing.Set[str],
) -> bool:
    """Implement check_parent."""
    # Check for $ref and allOf
    ref = schema.get("$ref")
    all_of = schema.get("allOf")
    if ref is None and all_of is None:
        return False

    # Handle $ref
    if ref is not None:
        if not isinstance(ref, str):
            raise exceptions.MalformedSchemaError("The value of $ref must be a string.")

        # Check for circular references
        if ref in seen_refs:
            raise exceptions.MalformedSchemaError(
                "Circular reference chain detected for the schema."
            )
        seen_refs.add(ref)

        ref_name, ref_schema = ref_helper.get_ref(ref=ref, schemas=schemas)

        # Check for name match base case
        if ref_name == parent_name:
            return schema_helper.constructable(schema=ref_schema, schemas=schemas)

        # Check referenced schema still inherits
        if schema_helper.inherits(schema=schema, schemas=schemas) is False:
            return False

        # Recursive case
        return _check_parent(ref_schema, parent_name, schemas, seen_refs)

    # Handle allOf
    if not isinstance(all_of, list):
        raise exceptions.MalformedSchemaError("The value of allOf must be a list.")
    return any(
        _check_parent(sub_schema, parent_name, schemas, seen_refs)
        for sub_schema in all_of
    )


def get_parent(*, schema: types.Schema, schemas: types.Schemas) -> str:
    """
    Get the name of the parent of the schema.

    Recursive function. The base cases are:
    1. the schema has $ref and the referenced schema is constructable.

    The recursive cases are:
    1. the schema has $ref where the referenced schema is not constructable in which
        case the function is called with the referenced schema and
    2. the schema has allOf in which case the return value of the first element that
        does not raise the MalformedSchemaError is returned.

    Raise MalformedSchemaError is the value of $ref is not a string.
    Raise MalformedSchemaError if the value of allOf is not a list.
    Raise MalformedSchemaError if the schema does not have $ref nor allOf.
    Raise MalformedSchemaError if the schema has allOf and all of the elements raise
        MalformedSchemaError.
    Raise MalformedSchemaError if a $ref value is seen again.

    Args:
        schema: The schema to retrieve the parent for.
        schemas: All the schemas.
        seen_refs: The $ref that have already been check to avoid circular dependencies.

    Returns:
        The name of the parent.

    """
    return _get_parent(schema, schemas, set())


def _get_parent(
    schema: types.Schema, schemas: types.Schemas, seen_refs: typing.Set[str]
) -> str:
    """Implement get_parent."""
    # Check for $ref and allOf
    ref = schema.get("$ref")
    all_of = schema.get("allOf")
    if ref is None and all_of is None:
        raise exceptions.MalformedSchemaError(
            "A schema that is marked as inhereting does not reference a valid parent."
        )

    # Handle $ref
    if ref is not None:
        if not isinstance(ref, str):
            raise exceptions.MalformedSchemaError("The value of $ref must be a string.")

        # Check for circular references
        if ref in seen_refs:
            raise exceptions.MalformedSchemaError(
                "Circular reference chain detected for the schema."
            )
        seen_refs.add(ref)

        ref_name, ref_schema = ref_helper.get_ref(ref=ref, schemas=schemas)

        # Check whether the referenced schema is constructible
        if schema_helper.constructable(schema=ref_schema, schemas=schemas):
            return ref_name

        # Recursive case
        return _get_parent(ref_schema, schemas, seen_refs)

    # Handle allOf
    if not isinstance(all_of, list):
        raise exceptions.MalformedSchemaError("The value of allOf must be a list.")
    # Find first constructable schema
    for sub_schema in all_of:
        try:
            return _get_parent(sub_schema, schemas, seen_refs)
        except exceptions.MalformedSchemaError:
            pass
    # None of the schemas in allOf are constructable
    raise exceptions.MalformedSchemaError(
        "A schema that is marked as inhereting does not reference a valid parent."
    )


def get_parents(
    *, schema: types.Schema, schemas: types.Schemas
) -> typing.Generator[str, None, None]:
    """
    Retrieve all parents of a schema.

    Recursive function. Base cases:
    1. schema without $ref nor allOf then nothing is returned.

    Recursive cases:
    1. $ref in which case values are yielded from the function called with the
        referenced schema and the referenced name is yielded and
    2. allOf in which case values are yielded from the function called with each element
        of allOf.

    Raise MalformedSchemaError is the value of $ref is not a string.
    Raise MalformedSchemaError if the value of allOf is not a list.
    Raise MalformedSchemaError if the same $ref value is seen again.

    Args:
        schema: The schema to retrieve all parents for.
        schemas: All the schemas.

    Returns:
        A generator with all parents of the schema.

    """
    return _get_parents(schema, schemas, set())


def _get_parents(
    schema: types.Schema, schemas: types.Schemas, seen_refs: typing.Set[str]
) -> typing.Generator[str, None, None]:
    """Implement get_parents."""
    # Check for $ref and allOf
    ref = schema.get("$ref")
    all_of = schema.get("allOf")

    # Handle $ref
    if ref is not None:
        if not isinstance(ref, str):
            raise exceptions.MalformedSchemaError("The value of $ref must be a string.")

        # Check for circular references
        if ref in seen_refs:
            raise exceptions.MalformedSchemaError(
                "Circular reference chain detected for the schema."
            )
        seen_refs.add(ref)

        ref_name, ref_schema = ref_helper.get_ref(ref=ref, schemas=schemas)

        # Check for inherits
        if schema_helper.inherits(schema=schema, schemas=schemas):
            yield from _get_parents(ref_schema, schemas, seen_refs)

        if schema_helper.constructable(schema=ref_schema, schemas=schemas):
            yield ref_name
        return

    # Handle allOf
    if all_of is not None:
        if not isinstance(all_of, list):
            raise exceptions.MalformedSchemaError("The value of allOf must be a list.")
        for sub_schema in all_of:
            yield from _get_parents(sub_schema, schemas, seen_refs)


def retrieve_parent(*, schema: types.Schema, schemas: types.Schemas) -> str:
    """
    Get or check the name of the parent.

    If x-inherits is True, get the name of the parent. If it is a string, check the
    parent.

    Raise InheritanceError if x-inherits is not defined or False.

    Args:
        schema: The schema to retrieve the parent for.
        schemas: All the schemas.

    Returns:
        The parent.

    """
    inherits = peek_helper.inherits(schema=schema, schemas=schemas)
    if inherits is True:
        return get_parent(schema=schema, schemas=schemas)
    if isinstance(inherits, str):
        if not check_parent(schema=schema, parent_name=inherits, schemas=schemas):
            raise exceptions.InheritanceError(
                f"The x-inherits value {inherits} is not a valid parent."
            )
        return inherits
    raise exceptions.InheritanceError(
        "Cannot retrieve the name of the parent if x-inherits is not defined or False."
    )


def retrieve_model_parents_schema(schema: types.Schema) -> types.Schema:
    """
    Retrieve the combined schema of the input and all its parent schemas.

    Args:
        schema: The model schema. Assume that it does not contain any $ref nor allOf.

    Returns:
        The combined schema.

    """
    schemas = _retrieve_model_parents_schema(schema)
    schema = {"allOf": schemas}
    return all_of_helper.merge(schema=schema, schemas={})


def _retrieve_model_parents_schema(
    schema: types.Schema,
) -> typing.Generator[types.Schema, None, None]:
    """
    Retrieve all the schemas of the inheritance chain, including the initial schema.

    Recursive function. Base case is schema without x-inherits where schema is yielded.
    Recursive case is schema with x-inherits where the function is called on the parent
    schema and then the schema is yielded.

    Args:
        schema: The model schema. Assume that it does not contain any $ref nor allOf.

    Returns:
        Generator with all the schemas.

    """
    inherits = ext_prop_helper.get(source=schema, name="x-inherits")
    if isinstance(inherits, str):
        parent_schema = facades.models.get_model_schema(name=inherits)
        if parent_schema is None:
            raise exceptions.InheritanceError(f"The parent {inherits} is not defined.")
        yield from _retrieve_model_parents_schema(parent_schema)
    yield schema


@enum.unique
class Type(str, enum.Enum):
    """The type of inheritance."""

    NONE = "NONE"
    JOINED_TABLE = "JOINED_TABLE"
    SINGLE_TABLE = "SINGLE_TABLE"


def calculate_type(*, schema: types.Schema, schemas: types.Schemas) -> Type:
    """
    Calculate the type of inheritance.

    Assume the schema and any parent schema is constructable and valid.

    The rules are:
    1. if the schema does not inherit return NONE,
    2. if the parent and child tablename are different return JOINED_TABLE and
    3. else return SINGLE_TABLE.

    Args:
        schema: The schema to calculate the type for.
        schemas: All defined schemas used to resolve any $ref.

    Returns:
        The type of inheritance.

    """
    if not schema_helper.inherits(schema=schema, schemas=schemas):
        return Type.NONE

    parent = retrieve_parent(schema=schema, schemas=schemas)
    parent_schema = schemas[parent]
    parent_tablename = peek_helper.tablename(schema=parent_schema, schemas=schemas)
    tablename = peek_helper.prefer_local(
        get_value=peek_helper.tablename, schema=schema, schemas=schemas
    )

    if parent_tablename == tablename:
        return Type.SINGLE_TABLE
    return Type.JOINED_TABLE



"""Retrieve artifacts for a simple property."""

from .... import helpers as oa_helpers
from .... import types as oa_types
from .. import types


def get(
    schemas: oa_types.Schemas, schema: oa_types.Schema, required: bool
) -> types.SimplePropertyArtifacts:
    """
    Retrieve the artifacts for a simple property.

    Args:
        schemas: All the defined schemas.
        schema: The schema of the simple property to gather artifacts for.
        required: WHether the property appears in the required list.

    Returns:
        The artifacts for the property.

    """
    type_ = oa_helpers.peek.type_(schema=schema, schemas=schemas)
    format_ = oa_helpers.peek.prefer_local(
        get_value=oa_helpers.peek.format_, schema=schema, schemas=schemas
    )
    max_length = oa_helpers.peek.prefer_local(
        get_value=oa_helpers.peek.max_length, schema=schema, schemas=schemas
    )
    nullable = oa_helpers.peek.prefer_local(
        get_value=oa_helpers.peek.nullable, schema=schema, schemas=schemas
    )

    description = oa_helpers.peek.prefer_local(
        get_value=oa_helpers.peek.description, schema=schema, schemas=schemas
    )

    default = oa_helpers.peek.prefer_local(
        get_value=oa_helpers.peek.default, schema=schema, schemas=schemas
    )
    server_default = oa_helpers.peek.prefer_local(
        get_value=oa_helpers.peek.server_default, schema=schema, schemas=schemas
    )

    read_only = oa_helpers.peek.prefer_local(
        get_value=oa_helpers.peek.read_only, schema=schema, schemas=schemas
    )
    write_only = oa_helpers.peek.prefer_local(
        get_value=oa_helpers.peek.write_only, schema=schema, schemas=schemas
    )

    primary_key = oa_helpers.peek.prefer_local(
        get_value=oa_helpers.peek.primary_key, schema=schema, schemas=schemas
    )
    autoincrement = oa_helpers.peek.prefer_local(
        get_value=oa_helpers.peek.autoincrement, schema=schema, schemas=schemas
    )
    index = oa_helpers.peek.prefer_local(
        get_value=oa_helpers.peek.index, schema=schema, schemas=schemas
    )
    unique = oa_helpers.peek.prefer_local(
        get_value=oa_helpers.peek.unique, schema=schema, schemas=schemas
    )

    foreign_key = oa_helpers.peek.prefer_local(
        get_value=oa_helpers.peek.foreign_key, schema=schema, schemas=schemas
    )

    kwargs = oa_helpers.peek.prefer_local(
        get_value=oa_helpers.peek.kwargs, schema=schema, schemas=schemas
    )
    foreign_key_kwargs = oa_helpers.peek.prefer_local(
        get_value=oa_helpers.peek.foreign_key_kwargs, schema=schema, schemas=schemas
    )

    dict_ignore = oa_helpers.peek.dict_ignore(schema=schema, schemas=schemas)

    # Generate the schema
    schema_artifact: oa_types.ColumnSchema = {
        "type": type_,
    }
    if format_ is not None:
        schema_artifact["format"] = format_
    if max_length is not None:
        schema_artifact["maxLength"] = max_length
    if description is not None:
        schema_artifact["description"] = description
    if nullable is not None:
        schema_artifact["nullable"] = nullable
    if default is not None:
        schema_artifact["default"] = default
    if read_only is not None:
        schema_artifact["readOnly"] = read_only
    if write_only is not None:
        schema_artifact["writeOnly"] = write_only
    if dict_ignore is not None:
        schema_artifact["x-dict-ignore"] = dict_ignore

    return types.SimplePropertyArtifacts(
        type=oa_types.PropertyType.SIMPLE,
        description=description,
        schema=schema_artifact,
        required=required,
        open_api=types.OpenApiSimplePropertyArtifacts(
            type=type_,
            format=format_,
            max_length=max_length,
            nullable=nullable,
            default=default,
            read_only=read_only,
            write_only=write_only,
        ),
        extension=types.ExtensionSimplePropertyArtifacts(
            primary_key=primary_key is True,
            autoincrement=autoincrement,
            index=index,
            unique=unique,
            server_default=server_default,
            foreign_key=foreign_key,
            kwargs=kwargs,
            foreign_key_kwargs=foreign_key_kwargs,
            dict_ignore=dict_ignore,
        ),
    )



"""Define validation rules for JSON properties."""

from .... import exceptions
from .... import helpers
from .... import types as oa_types
from .. import types
from . import simple


def check(schemas: oa_types.Schemas, schema: oa_types.Schema) -> types.Result:
    """
    Check the schema of a simple property (not an object nor an array).

    Args:
        schemas: The schemas used to resolve any $ref.
        schema: The schema of the property.

    Returns:
        A result with whether the schema is valid and a reason if it is not.

    """
    try:
        helpers.peek.prefer_local(
            get_value=helpers.peek.nullable, schema=schema, schemas=schemas
        )
        helpers.peek.prefer_local(
            get_value=helpers.peek.description, schema=schema, schemas=schemas
        )
        helpers.peek.prefer_local(
            get_value=helpers.peek.read_only, schema=schema, schemas=schemas
        )
        helpers.peek.prefer_local(
            get_value=helpers.peek.write_only, schema=schema, schemas=schemas
        )
        helpers.peek.prefer_local(
            get_value=helpers.peek.index, schema=schema, schemas=schemas
        )
        helpers.peek.prefer_local(
            get_value=helpers.peek.unique, schema=schema, schemas=schemas
        )
        helpers.peek.prefer_local(
            get_value=helpers.peek.primary_key, schema=schema, schemas=schemas
        )
        autoincrement = helpers.peek.peek_key(
            schema=schema, schemas=schemas, key="x-autoincrement"
        )
        if autoincrement is not None:
            return types.Result(False, "json properties do not support x-autoincrement")
        server_default = helpers.peek.peek_key(
            schema=schema, schemas=schemas, key="x-server-default"
        )
        if server_default is not None:
            return types.Result(
                False, "json properties do not support x-server-default"
            )
        # Checks kwargs, foreign key and foreign key kwargs
        kwargs_result = simple.check_kwargs(schema=schema, schemas=schemas)
        if kwargs_result is not None:
            return kwargs_result

    except exceptions.MalformedSchemaError as exc:
        return types.Result(False, f"malformed schema :: {exc}")
    except exceptions.SchemaNotFoundError as exc:
        return types.Result(False, f"reference :: {exc}")

    return types.Result(True, None)



"""Types for model generation."""

import dataclasses
import functools
import textwrap
import typing

from open_alchemy import types


@dataclasses.dataclass
class ColumnArtifacts:
    """Artifacts for the column portion of a model template."""

    # The name of the column based on the name of the property
    name: str
    # The type of the column
    type: str
    # The description of the column based on the description of the property
    description: typing.Optional[str] = None


@dataclasses.dataclass
class ColumnArgArtifacts:
    """Artifacts for the __init__ and from_dict signatures of a model template."""

    # The name of the argument
    name: str
    # The type of the argument for __init__
    init_type: str
    # The type of the argument for from_dict
    from_dict_type: str
    # The default value for the column
    default: types.TColumnDefault = None
    # Whether the column is read only
    read_only: typing.Optional[bool] = None


@dataclasses.dataclass
class ArgArtifacts:
    """Artifacts for the __init__ and from_dict args for a model."""

    # The artifacts for the arguments that are required
    required: typing.Iterable[ColumnArgArtifacts]
    # The artifacts for the arguments that are not required
    not_required: typing.Iterable[ColumnArgArtifacts]


@dataclasses.dataclass
class SQLAlchemyModelArtifacts:
    """Artifacts for the SQLAlchemy model."""

    # The name of the model based on the name of the schema
    name: str
    # Whether the columns are empty based on whether any properties are defined on the
    # schema and any back references
    empty: bool
    # The columns for the model based on the properties of the schema and any back
    # references
    columns: typing.List[ColumnArtifacts]
    # The artifacts for the arguments for __init__ and from_dict based on the properties
    # of the schema and any back references
    arg: ArgArtifacts
    # The parent class based on the version of Python
    parent_cls: str
    # The description of the model based on the description of the schema
    description: typing.Optional[str] = None

    # Calculated properties
    @property
    def docstring(self):
        """Calculate docstring for model."""
        return model_docstring(self)

    @property
    def init_docstring(self):
        """Calculate docstring for __init__ function of model."""
        return model_function_docstring(self, "Construct.")

    @property
    def from_dict_docstring(self):
        """Calculate docstring for from_dict function of model."""
        return model_function_docstring(
            self,
            "Construct from a dictionary (eg. a POST payload).",
            "Model instance based on the dictionary.",
        )


@dataclasses.dataclass
class TypedDictClassArtifacts:
    """Artifacts for a TypedDict class."""

    # The properties for the TypedDict
    props: typing.List[ColumnArtifacts]
    # Whether the properties list is empty
    empty: bool
    # The name of the TypedDict
    name: typing.Optional[str]
    # The name of the parent class
    parent_class: typing.Optional[str]


@dataclasses.dataclass
class TypedDictArtifacts:
    """Artifacts for the TypedDicts for a model."""

    # The artifacts for the required TypedDict
    required: TypedDictClassArtifacts
    # The artifacts for the required TypedDict
    not_required: TypedDictClassArtifacts


@dataclasses.dataclass
class ModelArtifacts:
    """Artifacts for a model template."""

    # The artifacts for the SQLAlchemy model
    sqlalchemy: SQLAlchemyModelArtifacts
    # The artifacts for the TypedDicts
    typed_dict: TypedDictArtifacts


_DocstringWrapper = textwrap.TextWrapper(width=75)  # pylint: disable=invalid-name
_DEFAULT_DOCSTRING = "SQLAlchemy model protocol."


def model_docstring(artifacts: SQLAlchemyModelArtifacts) -> str:
    """
    Create docstring for model from model artifacts.

    Args:
        artifacts: The artifacts of the model.

    Returns:
        The docstring for the model.

    """
    if artifacts.description is None and artifacts.empty:
        return _DEFAULT_DOCSTRING

    # Calculate description
    description: str
    if artifacts.description is None:
        description = f"""
    {_DEFAULT_DOCSTRING}"""
    else:
        wrapped_description = _DocstringWrapper.wrap(artifacts.description)
        joined_description = "\n    ".join(wrapped_description)
        description = f"""
    {_DEFAULT_DOCSTRING}

    {joined_description}"""

    # Calculate docs for the attributes
    attr_docs = ""
    if not artifacts.empty:
        model_attr_docs_model_name_set = functools.partial(
            model_attr_docs, model_name=artifacts.name
        )
        mapped_attrs = map(model_attr_docs_model_name_set, artifacts.columns)
        joined_attrs = "\n        ".join(mapped_attrs)
        attr_docs = f"""

    Attrs:
        {joined_attrs}"""

    return f"""{description}{attr_docs}

    """


def model_function_docstring(
    artifacts: SQLAlchemyModelArtifacts,
    function_description: str,
    return_value_description: typing.Optional[str] = None,
) -> str:
    """
    Create docstring for a function from model artifacts.

    Args:
        artifacts: The artifacts of the model.
        function_description: The description for the function.
        return_value_description: The description of the return value.

    Returns:
        The docstring for the model.

    """
    if artifacts.empty and return_value_description is None:
        return function_description

    # Calculate return value documentation
    return_value_doc = ""
    if return_value_description is not None:
        return_value_doc = f"""Returns:
            {return_value_description}

        """

    if artifacts.empty:
        return f"""
        {function_description}

        {return_value_doc}"""

    # Calculate docs for the arguments
    model_arg_docs_model_name_set = functools.partial(
        model_arg_docs, model_name=artifacts.name
    )
    mapped_args = map(model_arg_docs_model_name_set, artifacts.columns)
    joined_args = "\n            ".join(mapped_args)

    return f"""
        {function_description}

        Args:
            {joined_args}

        {return_value_doc}"""


_AttrFirstWrapper = textwrap.TextWrapper(width=71)  # pylint: disable=invalid-name
_AttrRemainingWrapper = textwrap.TextWrapper(width=67)  # pylint: disable=invalid-name


def model_attr_docs(artifacts: ColumnArtifacts, model_name: str) -> str:
    """
    Calculate attribute documentation.

    Args:
        artifacts: The artifacts for the column to produce attribute documentation for.
        model_name: The name of the model that contains the attribute.

    Returns:
        The documentation for the attribute.

    """
    # Calculating docs for the attribute
    doc = _calculate_column_doc(artifacts=artifacts, model_name=model_name)

    # Wrapping and joining
    return _wrap_column_doc(
        doc=doc,
        first_wrapper=_AttrFirstWrapper,
        remaining_wrapper=_AttrRemainingWrapper,
        line_separator="\n            ",
    )


_ArgFirstWrapper = textwrap.TextWrapper(width=67)  # pylint: disable=invalid-name
_ArgRemainingWrapper = textwrap.TextWrapper(width=63)  # pylint: disable=invalid-name


def model_arg_docs(artifacts: ColumnArtifacts, model_name: str) -> str:
    """
    Calculate argument documentation.

    Args:
        artifacts: The artifacts for the column to produce argument documentation for.
        model_name: The name of the model that contains the argument.

    Returns:
        The documentation for the argument.

    """
    # Calculating docs for the argument
    doc = _calculate_column_doc(artifacts=artifacts, model_name=model_name)

    # Wrapping and joining
    return _wrap_column_doc(
        doc=doc,
        first_wrapper=_ArgFirstWrapper,
        remaining_wrapper=_ArgRemainingWrapper,
        line_separator="\n                ",
    )


def _calculate_column_doc(artifacts: ColumnArtifacts, model_name: str) -> str:
    """
    Calculate the documentation for the column.

    Args:
        artifacts: The artifacts for the column.
        model_name: The name of the model to which the column belongs.

    Returns:
        The documentation for the column.

    """
    description: str
    if artifacts.description is not None:
        description = artifacts.description
    else:
        description = f"The {artifacts.name} of the {model_name}."
    return f"{artifacts.name}: {description}"


def _wrap_column_doc(
    *,
    doc: str,
    first_wrapper: textwrap.TextWrapper,
    remaining_wrapper: textwrap.TextWrapper,
    line_separator: str,
) -> str:
    """
    Calculate the documentation for a column.

    Args:
        doc: The documentation string for the column.
        first_wrapper: The text wrapper to use for the first line.
        remaining_wrapper: The wrapper to use for all remaining lines.
        line_separator: The string to use between lines.

    Returns:
        The documentation for the column.

    """
    # Wrapping and joining
    wrapped_doc = first_wrapper.wrap(doc)
    if len(wrapped_doc) > 1:
        remaining_doc = " ".join(wrapped_doc[1:])
        remaining_doc_wrapped = remaining_wrapper.wrap(remaining_doc)
        wrapped_doc = [wrapped_doc[0]] + remaining_doc_wrapped
    return line_separator.join(wrapped_doc)



"""Retrieve artifacts for backref property."""

import copy
import typing

from .... import helpers as oa_helpers
from .... import types as oa_types
from ... import helpers
from .. import types

OPEN_API_TO_SUB_TYPE: typing.Dict[
    str,
    typing.Union[
        oa_types.Literal[types.BackrefSubType.OBJECT],
        oa_types.Literal[types.BackrefSubType.ARRAY],
    ],
] = {
    "object": types.BackrefSubType.OBJECT,
    "array": types.BackrefSubType.ARRAY,
}


def get(
    schemas: oa_types.Schemas, schema: oa_types.Schema
) -> types.BackrefPropertyArtifacts:
    """
    Retrieve the artifacts for a backref property.

    Args:
        schemas: All the defined schemas.
        schema: The schema of the backref property to gather artifacts for.

    Returns:
        The artifacts for the property.

    """
    schema = copy.deepcopy(
        oa_helpers.schema.prepare_deep(schema=schema, schemas=schemas)
    )

    type_ = oa_helpers.peek.type_(schema=schema, schemas=schemas)
    assert type_ in OPEN_API_TO_SUB_TYPE
    sub_type = OPEN_API_TO_SUB_TYPE[type_]

    description = oa_helpers.peek.description(schema=schema, schemas=schemas)

    # Get property names
    properties_items: typing.Iterable[typing.Tuple[str, typing.Any]]
    if sub_type == types.BackrefSubType.OBJECT:  # noqa: E721
        properties_items = helpers.iterate.properties_items(
            schema=schema, schemas=schemas
        )
    else:
        items_schema = oa_helpers.peek.items(schema=schema, schemas=schemas)
        assert items_schema is not None
        properties_items = helpers.iterate.properties_items(
            schema=items_schema, schemas=schemas
        )
    properties_names = map(lambda args: args[0], properties_items)

    # Remove extension properties from schema
    helpers.clean.extension(schema=schema)
    if sub_type == types.BackrefSubType.ARRAY:  # noqa: E721
        helpers.clean.extension(schema=schema["items"])
        for property_schema in schema["items"]["properties"].values():
            helpers.clean.extension(schema=property_schema)
    else:
        for property_schema in schema["properties"].values():
            helpers.clean.extension(schema=property_schema)

    return types.BackrefPropertyArtifacts(
        type=oa_types.PropertyType.BACKREF,
        description=description,
        sub_type=sub_type,
        schema=schema,
        required=None,
        properties=list(properties_names),
    )



"""The artifact types."""

from ... import types

TMixins = types.TMixins
TKwargs = types.TKwargs
PropertyArtifacts = types.PropertyArtifacts
OpenApiSimplePropertyArtifacts = types.OpenApiSimplePropertyArtifacts
ExtensionSimplePropertyTypedDict = types.ExtensionSimplePropertyTypedDict
ExtensionSimplePropertyArtifacts = types.ExtensionSimplePropertyArtifacts
SimplePropertyTypedDict = types.SimplePropertyTypedDict
SimplePropertyArtifacts = types.SimplePropertyArtifacts
OpenApiJsonPropertyTypedDict = types.OpenApiJsonPropertyTypedDict
OpenApiJsonPropertyArtifacts = types.OpenApiJsonPropertyArtifacts
ExtensionJsonPropertyTypedDict = types.ExtensionJsonPropertyTypedDict
ExtensionJsonPropertyArtifacts = types.ExtensionJsonPropertyArtifacts
JsonPropertyTypedDict = types.JsonPropertyTypedDict
JsonPropertyArtifacts = types.JsonPropertyArtifacts
RelationshipPropertyArtifacts = types.RelationshipPropertyArtifacts
NotManyToManyRelationshipPropertyArtifacts = (
    types.NotManyToManyRelationshipPropertyArtifacts
)
OneToManyRelationshipPropertyTypedDict = types.OneToManyRelationshipPropertyTypedDict
OneToManyRelationshipPropertyArtifacts = types.OneToManyRelationshipPropertyArtifacts
XToOneRelationshipPropertyTypedDict = types.XToOneRelationshipPropertyTypedDict
XToOneRelationshipPropertyArtifacts = types.XToOneRelationshipPropertyArtifacts
ManyToOneRelationshipPropertyTypedDict = types.ManyToOneRelationshipPropertyTypedDict
ManyToOneRelationshipPropertyArtifacts = types.ManyToOneRelationshipPropertyArtifacts
OneToOneRelationshipPropertyTypedDict = types.OneToOneRelationshipPropertyTypedDict
OneToOneRelationshipPropertyArtifacts = types.OneToOneRelationshipPropertyArtifacts
ManyToManyRelationshipPropertyTypedDict = types.ManyToManyRelationshipPropertyTypedDict
ManyToManyRelationshipPropertyArtifacts = types.ManyToManyRelationshipPropertyArtifacts
TAnyRelationshipPropertyTypedDict = types.TAnyRelationshipPropertyTypedDict
TAnyRelationshipPropertyArtifacts = types.TAnyRelationshipPropertyArtifacts
BackrefPropertyTypedDict = types.BackrefPropertyTypedDict
BackrefSubType = types.BackrefSubType
BackrefPropertyArtifacts = types.BackrefPropertyArtifacts
TAnyPropertyTypedDict = types.TAnyPropertyTypedDict
TAnyPropertyArtifacts = types.TAnyPropertyArtifacts
PropertyValue = types.PropertyValue
PropertiesValue = types.PropertiesValue
ModelTypedDict = types.ModelTypedDict
ModelValue = types.ModelValue
ModelBackrefArtifacts = types.ModelBackrefArtifacts
ModelExPropertiesArtifacts = types.ModelExPropertiesArtifacts
ModelArtifacts = types.ModelArtifacts
ModelsValue = types.ModelsValue
SpecValue = types.SpecValue
ModelsModelArtifacts = types.ModelsModelArtifacts



"""Check that all x-secondary are unique."""

import typing

from ... import helpers as oa_helpers
from ... import types as oa_types
from .. import helpers
from . import types


class _TSeenSecondary(typing.NamedTuple):
    """Records information about the secondary that has been seen."""

    schema_name: str
    property_name: str


def check(*, schemas: oa_types.Schemas) -> types.Result:
    """
    Check that all x-secondary are unique.

    Assume that schemas are otherwise valid.

    The algorithm is:
    1. iterate over constructable schemas,
    2. iterate over properties restricted to model context,
    3. filter for many-to-many relationship properties,
    3. record the secondary, property and schema name,
    4. if a secondary has already been seen, return result that is not valid and
    5. return valid result.

    Args:
        schemas: The schemas to check.

    Returns:
        Whether the schemas have unique secondary values and the reason if not.

    """
    constructable_schemas = helpers.iterate.constructable(schemas=schemas)

    # Track information about the x-secondary values that are encountered
    seen_secondaries: typing.Dict[str, _TSeenSecondary] = {}

    for schema_name, schema in constructable_schemas:
        properties = helpers.iterate.properties_items(
            schema=schema, schemas=schemas, stay_within_model=True
        )

        for property_name, property_schema in properties:
            # Skip properties that are not many-to-many relationships
            is_many_to_many = oa_helpers.relationship.is_relationship_type(
                type_=oa_types.RelationshipType.MANY_TO_MANY,
                schema=property_schema,
                schemas=schemas,
            )
            if not is_many_to_many:
                continue

            # Retrieve secondary value
            secondary = helpers.association.get_secondary(
                schema=property_schema, schemas=schemas
            )

            # Check whether secondary has already been seen
            if secondary not in seen_secondaries:
                seen_secondaries[secondary] = _TSeenSecondary(
                    schema_name=schema_name, property_name=property_name
                )
                continue

            # Construct reason that schemas are not valid
            seen_on_schema = seen_secondaries[secondary]
            return types.Result(
                valid=False,
                reason=(
                    f'duplicate "x-secondary value {secondary} defined on the property '
                    f"{property_name} on the schema {schema_name} has already been "
                    f"defined on the property {seen_on_schema.property_name} on the "
                    f"schema {seen_on_schema.schema_name}"
                ),
            )

    return types.Result(valid=True, reason=None)



"""Performs operations on the schemas to prepare them for further processing."""

from .. import types as _types
from . import artifacts
from . import association
from . import backref
from . import foreign_key
from . import helpers
from . import validation


def process(*, schemas: _types.Schemas) -> None:
    """
    Pre-process schemas.

    The processing actions executed are:
    1. Calculate the back references.

    Args:
        schemas: The schemas to pre-process in place.

    """
    validation.process(schemas=schemas)
    backref.process(schemas=schemas)
    foreign_key.process(schemas=schemas)
    association.process(schemas=schemas)



"""Assemble the final schema and return its type."""

import typing

from open_alchemy import exceptions
from open_alchemy import facades
from open_alchemy import types

from . import ext_prop as ext_prop_helper
from . import ref as ref_helper


class PeekValue(types.Protocol):
    """Defines interface for peek functions."""

    def __call__(self, *, schema: types.Schema, schemas: types.Schemas) -> typing.Any:
        """Call signature for peek functions."""
        ...


def type_(*, schema: types.Schema, schemas: types.Schemas) -> str:
    """
    Get the type of the schema.

    Raises TypeMissingError if the final schema does not have a type or the value is
    not a string.

    Args:
        schema: The schema for which to get the type.
        schemas: The schemas for $ref lookup.

    Returns:
        The type of the schema.

    """
    value = peek_key(schema=schema, schemas=schemas, key="type")
    if value is None:
        raise exceptions.TypeMissingError("Every property requires a type.")
    if not isinstance(value, str):
        raise exceptions.TypeMissingError(
            "A type property value must be of type string."
        )
    return value


def nullable(*, schema: types.Schema, schemas: types.Schemas) -> typing.Optional[bool]:
    """
    Retrieve the nullable property from a property schema.

    Raises MalformedSchemaError if the nullable value is not a boolean.

    Args:
        schema: The schema to get the nullable from.
        schemas: The schemas for $ref lookup.

    Returns:
        The nullable value.

    """
    value = peek_key(schema=schema, schemas=schemas, key="nullable")
    if value is None:
        return None
    if not isinstance(value, bool):
        raise exceptions.MalformedSchemaError(
            "A nullable value must be of type boolean."
        )
    return value


def format_(*, schema: types.Schema, schemas: types.Schemas) -> typing.Optional[str]:
    """
    Retrieve the format property from a property schema.

    Raises MalformedSchemaError if the format value is not a string.

    Args:
        schema: The schema to get the format from.
        schemas: The schemas for $ref lookup.

    Returns:
        The format value or None if it was not found.

    """
    value = peek_key(schema=schema, schemas=schemas, key="format")
    if value is None:
        return None
    if not isinstance(value, str):
        raise exceptions.MalformedSchemaError("A format value must be of type string.")
    return value


def autoincrement(
    *, schema: types.Schema, schemas: types.Schemas
) -> typing.Optional[bool]:
    """
    Retrieve the x-autoincrement property from a property schema.

    Raises MalformedSchemaError if the x-autoincrement value is not a boolean.

    Args:
        schema: The schema to get the x-autoincrement from.
        schemas: The schemas for $ref lookup.

    Returns:
        The x-autoincrement value.

    """
    value = peek_key(schema=schema, schemas=schemas, key="x-autoincrement")
    if value is None:
        return None
    if not isinstance(value, bool):
        raise exceptions.MalformedSchemaError(
            "A autoincrement value must be of type boolean."
        )
    return value


def index(*, schema: types.Schema, schemas: types.Schemas) -> typing.Optional[bool]:
    """
    Retrieve the x-index property from a property schema.

    Raises MalformedSchemaError if the x-index value is not a boolean.

    Args:
        schema: The schema to get the x-index from.
        schemas: The schemas for $ref lookup.

    Returns:
        The x-index value.

    """
    value = peek_key(schema=schema, schemas=schemas, key="x-index")
    if value is None:
        return None
    if not isinstance(value, bool):
        raise exceptions.MalformedSchemaError("A index value must be of type boolean.")
    return value


def unique(*, schema: types.Schema, schemas: types.Schemas) -> typing.Optional[bool]:
    """
    Retrieve the x-unique property from a property schema.

    Raises MalformedSchemaError if the x-unique value is not a boolean.

    Args:
        schema: The schema to get the x-unique from.
        schemas: The schemas for $ref lookup.

    Returns:
        The x-unique value.

    """
    value = peek_key(schema=schema, schemas=schemas, key="x-unique")
    if value is None:
        return None
    if not isinstance(value, bool):
        raise exceptions.MalformedSchemaError("A unique value must be of type boolean.")
    return value


def max_length(*, schema: types.Schema, schemas: types.Schemas) -> typing.Optional[int]:
    """
    Retrieve the maxLength property from a property schema.

    Raises MalformedSchemaError if the maxLength value is not an integer.

    Args:
        schema: The schema to get the maxLength from.
        schemas: The schemas for $ref lookup.

    Returns:
        The maxLength value or None if it was not found.

    """
    value = peek_key(schema=schema, schemas=schemas, key="maxLength")
    if value is None:
        return None
    if not isinstance(value, int) or isinstance(value, bool):
        raise exceptions.MalformedSchemaError(
            "A maxLength value must be of type integer."
        )
    return value


def read_only(*, schema: types.Schema, schemas: types.Schemas) -> typing.Optional[bool]:
    """
    Determine whether schema is readOnly.

    Raises MalformedSchemaError if the readOnly value is not a boolean.

    Args:
        schema: The schema to get readOnly from.
        schemas: The schemas for $ref lookup.

    Returns:
        Whether the schema is readOnly.

    """
    value = peek_key(schema=schema, schemas=schemas, key="readOnly")
    if value is None:
        return None
    if not isinstance(value, bool):
        raise exceptions.MalformedSchemaError(
            "A readOnly property must be of type boolean."
        )
    return value


def write_only(
    *, schema: types.Schema, schemas: types.Schemas
) -> typing.Optional[bool]:
    """
    Determine whether schema is writeOnly.

    Raises MalformedSchemaError if the writeOnly value is not a boolean.

    Args:
        schema: The schema to get writeOnly from.
        schemas: The schemas for $ref lookup.

    Returns:
        Whether the schema is writeOnly.

    """
    value = peek_key(schema=schema, schemas=schemas, key="writeOnly")
    if value is None:
        return None
    if not isinstance(value, bool):
        raise exceptions.MalformedSchemaError(
            "A writeOnly property must be of type boolean."
        )
    return value


def description(
    *, schema: types.Schema, schemas: types.Schemas
) -> typing.Optional[str]:
    """
    Retrieve the description property from a property schema.

    Raises MalformedSchemaError if the description value is not a string.

    Args:
        schema: The schema to get the description from.
        schemas: The schemas for $ref lookup.

    Returns:
        The description value or None if it was not found.

    """
    value = peek_key(schema=schema, schemas=schemas, key="description")
    if value is None:
        return None
    if not isinstance(value, str):
        raise exceptions.MalformedSchemaError(
            "A description value must be of type string."
        )
    return value


def primary_key(
    *, schema: types.Schema, schemas: types.Schemas
) -> typing.Optional[bool]:
    """
    Determine whether property schema is for a primary key.

    Raises MalformedSchemaError if the x-primary-key value is not a boolean.

    Args:
        schema: The schema to get x-primary-key from.
        schemas: The schemas for $ref lookup.

    Returns:
        Whether the schema is for a primary key property.

    """
    value = peek_key(schema=schema, schemas=schemas, key="x-primary-key")
    if value is None:
        return None
    if not isinstance(value, bool):
        raise exceptions.MalformedSchemaError(
            "The x-primary-key property must be of type boolean."
        )
    return value


def tablename(*, schema: types.Schema, schemas: types.Schemas) -> typing.Optional[str]:
    """
    Retrieve the x-tablename of the schema.

    Raises MalformedSchemaError if the x-tablename value is not a string.

    Args:
        schema: The schema to get x-tablename from.
        schemas: The schemas for $ref lookup.

    Returns:
        The x-tablename or None.

    """
    value = peek_key(schema=schema, schemas=schemas, key="x-tablename")
    if value is None:
        return None
    if not isinstance(value, str):
        raise exceptions.MalformedSchemaError(
            "The x-tablename property must be of type string."
        )
    return value


def inherits(
    *, schema: types.Schema, schemas: types.Schemas
) -> typing.Optional[typing.Union[str, bool]]:
    """
    Retrieve the value of the x-inherits extension property of the schema.

    Raises MalformedSchemaError if the value is not a string nor a boolean.

    Args:
        schema: The schema to get x-inherits from.
        schemas: The schemas for $ref lookup.

    Returns:
        The inherits or None.

    """
    value = peek_key(schema=schema, schemas=schemas, key="x-inherits")
    if value is None:
        return None
    if not isinstance(value, (str, bool)):
        raise exceptions.MalformedSchemaError(
            "The x-inherits property must be of type string or boolean."
        )
    return value


def json(*, schema: types.Schema, schemas: types.Schemas) -> typing.Optional[bool]:
    """
    Retrieve the value of the x-json extension property of the schema.

    Raises MalformedSchemaError if the value is not a boolean.

    Args:
        schema: The schema to get x-json from.
        schemas: The schemas for $ref lookup.

    Returns:
        The x-json value or None if the schema does not have the key.

    """
    value = peek_key(schema=schema, schemas=schemas, key="x-json")
    if value is None:
        return None
    if not isinstance(value, bool):
        raise exceptions.MalformedSchemaError(
            "The x-json property must be of type boolean."
        )
    return value


def backref(*, schema: types.Schema, schemas: types.Schemas) -> typing.Optional[str]:
    """
    Retrieve the x-backref of the schema.

    Raises MalformedSchemaError if the x-backref value is not a string.

    Args:
        schema: The schema to get x-backref from.
        schemas: The schemas for $ref lookup.

    Returns:
        The x-backref or None.

    """
    value = peek_key(schema=schema, schemas=schemas, key="x-backref")
    if value is None:
        return None
    if not isinstance(value, str):
        raise exceptions.MalformedSchemaError(
            "The x-backref property must be of type string."
        )
    return value


def secondary(*, schema: types.Schema, schemas: types.Schemas) -> typing.Optional[str]:
    """
    Retrieve the x-secondary of the schema.

    Raises MalformedSchemaError if the x-secondary value is not a string.

    Args:
        schema: The schema to get x-secondary from.
        schemas: The schemas for $ref lookup.

    Returns:
        The x-secondary or None.

    """
    value = peek_key(schema=schema, schemas=schemas, key="x-secondary")
    if value is None:
        return None
    if not isinstance(value, str):
        raise exceptions.MalformedSchemaError(
            "The x-secondary property must be of type string."
        )
    return value


def uselist(*, schema: types.Schema, schemas: types.Schemas) -> typing.Optional[bool]:
    """
    Retrieve the x-uselist of the schema.

    Raises MalformedSchemaError if the x-uselist value is not a boolean.

    Args:
        schema: The schema to get x-uselist from.
        schemas: The schemas for $ref lookup.

    Returns:
        The x-uselist or None.

    """
    value = peek_key(schema=schema, schemas=schemas, key="x-uselist")
    if value is None:
        return None
    if not isinstance(value, bool):
        raise exceptions.MalformedSchemaError(
            "The x-uselist property must be of type boolean."
        )
    return value


def items(*, schema: types.Schema, schemas: types.Schemas) -> typing.Optional[dict]:
    """
    Retrieve the items of the schema.

    Raises MalformedSchemaError if the items value is not a dictionary.

    Args:
        schema: The schema to get items from.
        schemas: The schemas for $ref lookup.

    Returns:
        The items or None.

    """
    value = peek_key(schema=schema, schemas=schemas, key="items")
    if value is None:
        return None
    if not isinstance(value, dict):
        raise exceptions.MalformedSchemaError(
            "The items property must be of type dict."
        )
    return value


def _check_kwargs(*, value: typing.Any, key: str) -> typing.Dict[str, typing.Any]:
    """Check the kwargs value."""
    # Check value
    if not isinstance(value, dict):
        raise exceptions.MalformedSchemaError(
            f"The {key} property must be of type dict."
        )
    # Check keys
    not_str_keys = filter(lambda key: not isinstance(key, str), value.keys())
    if next(not_str_keys, None) is not None:
        raise exceptions.MalformedSchemaError(
            f"The {key} property must have string keys."
        )
    return value


def kwargs(
    *, schema: types.Schema, schemas: types.Schemas
) -> typing.Optional[typing.Dict[str, typing.Any]]:
    """
    Retrieve the x-kwargs of the schema.

    Raises MalformedSchemaError if the x-kwargs value is not a dictionary.

    Args:
        schema: The schema to get x-kwargs from.
        schemas: The schemas for $ref lookup.

    Returns:
        The x-kwargs or None.

    """
    key = "x-kwargs"
    value = peek_key(schema=schema, schemas=schemas, key=key)
    if value is None:
        return None
    # Check value
    return _check_kwargs(value=value, key=key)


def foreign_key_kwargs(
    *, schema: types.Schema, schemas: types.Schemas
) -> typing.Optional[dict]:
    """
    Retrieve the x-foreign-key-kwargs of the schema.

    Raises MalformedSchemaError if the x-foreign-key-kwargs value is not a dictionary.

    Args:
        schema: The schema to get x-foreign-key-kwargs from.
        schemas: The schemas for $ref lookup.

    Returns:
        The x-foreign-key-kwargs or None.

    """
    key = "x-foreign-key-kwargs"
    value = peek_key(schema=schema, schemas=schemas, key=key)
    if value is None:
        return None
    # Check value
    return _check_kwargs(value=value, key=key)


def ref(*, schema: types.Schema, schemas: types.Schemas) -> typing.Optional[str]:
    """
    Retrieve the $ref of the schema.

    Raises MalformedSchemaError if the $ref value is not a dictionary.

    Args:
        schema: The schema to get $ref from.
        schemas: The schemas for $ref lookup.

    Returns:
        The $ref or None.

    """
    value = peek_key(schema=schema, schemas=schemas, key="$ref")
    if value is None:
        return None
    if not isinstance(value, str):
        raise exceptions.MalformedSchemaError(
            "The $ref property must be of type string."
        )
    return value


def foreign_key(
    *, schema: types.Schema, schemas: types.Schemas
) -> typing.Optional[str]:
    """
    Retrieve the x-foreign-key of the schema.

    Raises MalformedSchemaError if the x-foreign-key value is not a string.

    Args:
        schema: The schema to get x-foreign-key from.
        schemas: The schemas for $ref lookup.

    Returns:
        The x-foreign-key or None.

    """
    value = peek_key(schema=schema, schemas=schemas, key="x-foreign-key")
    if value is None:
        return None
    if not isinstance(value, str):
        raise exceptions.MalformedSchemaError(
            "The x-foreign-key property must be of type string."
        )
    return value


def foreign_key_column(
    *, schema: types.Schema, schemas: types.Schemas
) -> typing.Optional[str]:
    """
    Retrieve the x-foreign-key-column of the schema.

    Raises MalformedSchemaError if the x-foreign-key-column value is not a string.

    Args:
        schema: The schema to get x-foreign-key-column from.
        schemas: The schemas for $ref lookup.

    Returns:
        The x-foreign-key-column or None.

    """
    value = peek_key(schema=schema, schemas=schemas, key="x-foreign-key-column")
    if value is None:
        return None
    if not isinstance(value, str):
        raise exceptions.MalformedSchemaError(
            "The x-foreign-key-column property must be of type string."
        )
    return value


def composite_index(
    *, schema: types.Schema, schemas: types.Schemas
) -> typing.Optional[typing.List]:
    """
    Retrieve the x-composite-index of the schema.

    Args:
        schema: The schema to get x-composite-index from.
        schemas: The schemas for $ref lookup.

    Returns:
        The x-composite-index or None.

    """
    key = "x-composite-index"
    value = peek_key(schema=schema, schemas=schemas, key=key)
    if value is None:
        return None
    # Check value
    ext_prop_helper.get(source={key: value}, name=key)  # type: ignore
    return value


def composite_unique(
    *, schema: types.Schema, schemas: types.Schemas
) -> typing.Optional[typing.List]:
    """
    Retrieve the x-composite-unique of the schema.

    Args:
        schema: The schema to get x-composite-unique from.
        schemas: The schemas for $ref lookup.

    Returns:
        The x-composite-unique or None.

    """
    key = "x-composite-unique"
    value = peek_key(schema=schema, schemas=schemas, key=key)
    if value is None:
        return None
    # Check value
    ext_prop_helper.get(source={key: value}, name=key)  # type: ignore
    return value


def default(*, schema: types.Schema, schemas: types.Schemas) -> types.TColumnDefault:
    """
    Retrieve the default value and check it against the schema.

    Raises MalformedSchemaError if the default value does not conform with the schema.

    Args:
        schema: The schema to retrieve the default value from.

    Returns:
        The default or None.

    """
    # Retrieve value
    value = peek_key(schema=schema, schemas=schemas, key="default")
    if value is None:
        return None
    # Assemble schema
    resolved_schema: types.ColumnSchema = {
        "type": type_(schema=schema, schemas=schemas)
    }
    format_value = format_(schema=schema, schemas=schemas)
    max_length_value = max_length(schema=schema, schemas=schemas)
    if format_value is not None:
        resolved_schema["format"] = format_value
    if max_length_value is not None:
        resolved_schema["maxLength"] = max_length_value
    try:
        facades.jsonschema.validate(value, resolved_schema)
    except facades.jsonschema.ValidationError as exc:
        raise exceptions.MalformedSchemaError(
            "The default value does not conform to the schema. "
            f"The value is: {repr(value)}"
        ) from exc
    return value


def server_default(
    *, schema: types.Schema, schemas: types.Schemas
) -> typing.Optional[str]:
    """
    Retrieve the x-server-default property from a property schema.

    Raises MalformedSchemaError if the x-server-default value is not a string.

    Args:
        schema: The schema to get the x-server-default from.
        schemas: The schemas for $ref lookup.

    Returns:
        The x-server-default value.

    """
    value = peek_key(schema=schema, schemas=schemas, key="x-server-default")
    if value is None:
        return None
    if not isinstance(value, str):
        raise exceptions.MalformedSchemaError(
            "A x-server-default value must be of type string."
        )
    return value


def mixins(
    *, schema: types.Schema, schemas: types.Schemas
) -> typing.Optional[typing.List[str]]:
    """
    Retrieve the x-mixins of the schema.

    Args:
        schema: The schema to get x-mixins from.
        schemas: The schemas for $ref lookup.

    Returns:
        The x-mixins or None.

    """
    key = "x-mixins"
    value = peek_key(schema=schema, schemas=schemas, key=key)
    if value is None:
        return None

    # Check value
    ext_prop_helper.get(source={key: value}, name=key)  # type: ignore

    # Transform string to list
    if isinstance(value, str):
        value = [value]

    # Check that each value is a valid dot-separated identifier
    def valid(mixin_value: str) -> bool:
        """Check whether a mixin value is valid."""
        components = mixin_value.split(".")
        if len(components) < 2:
            return False

        invalid_components = map(
            lambda component: not component.isidentifier(), components
        )
        return not any(invalid_components)

    values_valid = map(lambda mixin_value: (mixin_value, valid(mixin_value)), value)
    first_invalid_value = next(filter(lambda args: not args[1], values_valid), None)
    if first_invalid_value is not None:
        raise exceptions.MalformedExtensionPropertyError(
            f'mixin values must be a valid import path, "{first_invalid_value[0]}" is '
            "not"
        )

    return value


def dict_ignore(
    *, schema: types.Schema, schemas: types.Schemas
) -> typing.Optional[bool]:
    """
    Retrieve the x-dict-ignore property from a property schema.

    Raises MalformedSchemaError if the x-dict-ignore value is not a boolean.

    Args:
        schema: The schema to get the x-dict-ignore from.
        schemas: The schemas for $ref lookup.

    Returns:
        The x-dict-ignore value.

    """
    value = peek_key(schema=schema, schemas=schemas, key="x-dict-ignore")
    if value is None:
        return None
    if not isinstance(value, bool):
        raise exceptions.MalformedSchemaError(
            "A x-dict-ignore value must be of type boolean."
        )
    return value


def peek_key(
    *,
    schema: types.Schema,
    schemas: types.Schemas,
    key: str,
    skip_ref: typing.Optional[str] = None,
) -> typing.Any:
    """
    Recursive type lookup.

    Raise MalformedSchemaError of a $ref value is seen again.

    Args:
        schema: The schema to look up the key in.
        schemas: All the schemas to resolve any $ref.
        key: The key to check for.
        seen_refs: All the $ref that have already been seen.
        skip_ref: The name of a reference to not follow.

    Returns:
        The key value (if found) or None.

    """
    return _peek_key(schema, schemas, key, set(), skip_ref=skip_ref)


def _check_schema_schemas_dict(schema: types.Schema, schemas: types.Schemas) -> None:
    """Check that schema and schemas are dict."""
    # Check schema and schemas are dict
    if not isinstance(schema, dict):
        raise exceptions.MalformedSchemaError("The schema must be a dictionary.")
    if not isinstance(schemas, dict):
        raise exceptions.MalformedSchemaError("The schemas must be a dictionary.")


def _check_ref_string(ref_value: typing.Any) -> str:
    """Check that value of $ref is string."""
    if not isinstance(ref_value, str):
        raise exceptions.MalformedSchemaError("The value of $ref must be a string.")
    return ref_value


def _check_circular_ref(ref_value: str, seen_refs: typing.Set[str]) -> None:
    """Check whether ref has ever been seen."""
    if ref_value in seen_refs:
        raise exceptions.MalformedSchemaError("Circular reference detected.")
    seen_refs.add(ref_value)


def _check_all_of_list(all_of: typing.Any) -> list:
    """Check that value of allOf is a list."""
    if not isinstance(all_of, list):
        raise exceptions.MalformedSchemaError("The value of allOf must be a list.")
    return all_of


def _check_sub_schema_dict(sub_schema: typing.Any) -> dict:
    """Check that a sub schema in an allOf is a dict."""
    if not isinstance(sub_schema, dict):
        raise exceptions.MalformedSchemaError(
            "The elements of allOf must be dictionaries."
        )
    return sub_schema


def _peek_key(
    schema: types.Schema,
    schemas: types.Schemas,
    key: str,
    seen_refs: typing.Set[str],
    skip_ref: typing.Optional[str],
) -> typing.Any:
    """Execute peek_key."""
    _check_schema_schemas_dict(schema, schemas)

    # Base case, look for type key
    value = schema.get(key)
    if value is not None:
        return value

    # Recursive case, look for $ref
    ref_value = schema.get("$ref")
    if ref_value is not None:
        ref_value_str = _check_ref_string(ref_value)
        _check_circular_ref(ref_value_str, seen_refs)

        ref_name, ref_schema = ref_helper.get_ref(ref=ref_value_str, schemas=schemas)
        if skip_ref is not None and ref_name == skip_ref:
            return None
        return _peek_key(ref_schema, schemas, key, seen_refs, skip_ref)

    # Recursive case, look for allOf
    all_of = schema.get("allOf")
    if all_of is not None:
        all_of_list = _check_all_of_list(all_of)
        for sub_schema in all_of_list:
            sub_schema_dict = _check_sub_schema_dict(sub_schema)
            value = _peek_key(sub_schema_dict, schemas, key, seen_refs, skip_ref)
            if value is not None:
                return value

    # Base case, type or ref not found or no type in allOf
    return None


def prefer_local(
    *, get_value: PeekValue, schema: types.Schema, schemas: types.Schemas
) -> typing.Any:
    """
    Retrieve the value using a function preferably without having to follow a $ref.

    1. Check for allOf:
        if found, iterate over schemas in allOf and skip any that contain $ref and
            return the value returned by get_value if it is not None.
    2. Return output of get_value called on the schema.

    Args:
        get_value: The function that knows how to retrieve the value.
        schema: The schema to process.
        schemas: All the schemas.

    Returns:
        The value returned by get_value preferably without following any $ref.

    """
    return _prefer_local(get_value, schema, schemas, set())


def _prefer_local(
    get_value: PeekValue,
    schema: types.Schema,
    schemas: types.Schemas,
    seen_refs: typing.Set[str],
) -> typing.Any:
    """Execute prefer_local."""
    _check_schema_schemas_dict(schema, schemas)

    # Handle $ref
    ref_value = schema.get("$ref")
    if ref_value is not None:
        ref_value_str = _check_ref_string(ref_value)
        _check_circular_ref(ref_value_str, seen_refs)

        _, ref_schema = ref_helper.get_ref(ref=ref_value_str, schemas=schemas)
        return _prefer_local(get_value, ref_schema, schemas, seen_refs)

    # Handle allOf
    all_of = schema.get("allOf")
    if all_of is not None:
        all_of_list = _check_all_of_list(all_of)
        all_of_list_dict = map(_check_sub_schema_dict, all_of_list)
        # Order putting any $ref last
        sorted_all_of = sorted(
            all_of_list_dict, key=lambda sub_schema: sub_schema.get("$ref") is not None
        )

        def map_to_value(sub_schema: types.Schema) -> typing.Any:
            """Use get_value to turn the schema into the value."""
            return _prefer_local(get_value, sub_schema, schemas, seen_refs)

        retrieved_values = map(map_to_value, sorted_all_of)
        not_none_retrieved_values = filter(
            lambda value: value is not None, retrieved_values
        )
        retrieved_value = next(not_none_retrieved_values, None)
        return retrieved_value

    return get_value(schema=schema, schemas=schemas)



"""Generate columns based on OpenAPI schema property."""

from open_alchemy import types as oa_types

from . import json
from . import relationship
from . import simple
from . import types


def column_factory(*, artifacts: oa_types.TAnyPropertyArtifacts) -> types.TReturnValue:
    """
    Generate column based on OpenAPI schema property.

    Args:
        artifacts: The artifacts for the column.

    Returns:
        The column, relationship or None if the property does not need it.

    """
    if artifacts.type == oa_types.PropertyType.SIMPLE:
        return simple.handle(artifacts=artifacts)
    if artifacts.type == oa_types.PropertyType.JSON:
        return json.handle(artifacts=artifacts)
    if artifacts.type == oa_types.PropertyType.BACKREF:
        return None
    return relationship.handle(artifacts=artifacts)



"""Used to resolve schema references."""

import functools
import json
import os
import re
import typing
from urllib import error
from urllib import request

from open_alchemy import exceptions
from open_alchemy import types

_REF_PATTER = re.compile(r"^#\/components\/schemas\/(\w+)$")


NameSchema = typing.Tuple[str, types.Schema]


def resolve(
    *,
    name: str,
    schema: types.Schema,
    schemas: types.Schemas,
    skip_name: typing.Optional[str] = None,
) -> NameSchema:
    """
    Resolve reference to another schema.

    Recursively resolves $ref until $ref key is no longer found. On each step, the name
    of the schema is recorded.

    Raises SchemaNotFoundError is a $ref resolution fails.
    Raise MalformedSchemaError of a $ref value is seen again.

    Args:
        name: The name of the schema from the last step.
        schema: The specification of the schema from the last step.
        schemas: Dictionary with all defined schemas used to resolve $ref.
        skip_name (optional): Skip the schema and return an empty schema instead.

    Returns:
        The first schema that no longer has the $ref key and the name of that schema.

    """
    return _resolve(name, schema, schemas, set(), skip_name)


def _resolve(
    name: str,
    schema: types.Schema,
    schemas: types.Schemas,
    seen_refs: typing.Set[str],
    skip_name: typing.Optional[str],
) -> NameSchema:
    """Implement resolve."""
    # Checking whether schema is a reference schema
    ref = schema.get("$ref")
    if ref is None:
        return name, schema
    # Check that ref is string
    if not isinstance(ref, str):
        raise exceptions.MalformedSchemaError("The value of $ref must be a string.")

    # Check for circular $ref
    if ref in seen_refs:
        raise exceptions.MalformedSchemaError("Circular reference chain detected.")
    seen_refs.add(ref)

    ref_name, ref_schema = get_ref(ref=ref, schemas=schemas)

    # Check if schema should be skipped
    if ref_name == skip_name:
        return name, {}

    return _resolve(ref_name, ref_schema, schemas, seen_refs, skip_name)


def get_ref(*, ref: str, schemas: types.Schemas) -> NameSchema:
    """
    Get the schema referenced by ref.

    Raises SchemaNotFoundError if a $ref resolution fails.

    Args:
        ref: The reference to the schema.
        schemas: The schemas to use to resolve the ref.

    Returns:
        The schema referenced by ref.

    """
    # Check for remote $ref
    if not ref.startswith("#"):
        return get_remote_ref(ref=ref)

    # Checking value of $ref
    match = _REF_PATTER.match(ref)
    if not match:
        raise exceptions.SchemaNotFoundError(
            f"{ref} format incorrect, expected #/components/schemas/<SchemaName>"
        )

    # Retrieving new schema
    ref_name = match.group(1)
    ref_schema = schemas.get(ref_name)
    if ref_schema is None:
        raise exceptions.SchemaNotFoundError(f"{ref_name} was not found in schemas.")

    return ref_name, ref_schema


# URL $ref regex
_URL_REF_PATTERN = re.compile(r"^(https?:)\/\/", re.IGNORECASE)


def _norm_context(*, context: str) -> str:
    """
    Normalize the path and case of a context.

    Args:
        context: The context to normalize.

    Returns:
        The normalized context.

    """
    if _URL_REF_PATTERN.search(context) is not None:
        return context
    norm_context = os.path.normpath(context)
    return os.path.normcase(norm_context)


def _separate_context_path(*, ref: str) -> typing.Tuple[str, str]:
    """
    Separate the context and path of a reference.

    Raise MalformedSchemaError if the reference does not contain #.

    Args:
        ref: The reference to separate.

    Returns:
        The context and path to the schema as a tuple.

    """
    try:
        ref_context, ref_schema = ref.split("#")
    except ValueError as exc:
        raise exceptions.MalformedSchemaError(
            f"A reference must contain exactly one #. Actual reference: {ref}"
        ) from exc
    return ref_context, ref_schema


# Regex for capturing the hostname and path from a URL
_HOSTNAME_REF_PATTERM = re.compile(r"^(https?:\/\/.*?)(\/.*)$", re.IGNORECASE)


def _add_remote_context(*, context: str, ref: str) -> str:
    """
    Add remote context to any $ref within a schema retrieved from a remote reference.

    There are 5 cases:
    1. The $ref value starts with # in which case the context is prepended.
    2. The $ref starts with a filename in which case only the directory portion of the
        context is prepended.
    3. The $ref starts with a relative path and ends with a file in which case the
        directory portion of the context is prepended and merged so that the shortest
        possible relative path is used.
    4. The $ref starts with a HTTP protocol, in which case no changes are made.
    5. The $ref starts with // in which case the HTTP protocol of the context is
        prepended.

    Raise SchemaNotFoundError if the $ref starts with // when the context does not start
        with a HTTP protocol.

    After the paths are merged the following operations are done:
    1. a normalized relative path is calculated (eg. turning ./dir1/../dir2 to ./dir2)
        and
    2. the case is normalized.

    Args:
        context: The context of the document from which the schema was retrieved which
            is the relative path to the file on the system from the base OpenAPI
            specification.
        ref: The value of a $ref within the schema.

    Returns:
        The $ref value with the context of the document included.

    """
    # Check for URL reference
    url_match = _URL_REF_PATTERN.search(ref)
    if url_match is not None:
        return ref
    if ref.startswith("//"):
        context_protocol = _URL_REF_PATTERN.search(context)
        if context_protocol is None:
            raise exceptions.SchemaNotFoundError(
                "A reference starting with // is only valid from within a document "
                f"loaded from a URL. The reference is {ref}, the location of the "
                f"document with the reference is {context}."
            )
        return f"{context_protocol.group(1)}{ref}"

    # Handle reference within document
    ref_context, ref_schema = _separate_context_path(ref=ref)
    if not ref_context:
        return f"{context}{ref}"

    # Break context into components
    # Default where context is not a URL
    context_hostname = ""
    context_path = context
    # Gather components if the context is a URL
    hostname_match = _HOSTNAME_REF_PATTERM.search(context)
    if hostname_match is not None:
        context_hostname = hostname_match.group(1)
        context_path = hostname_match.group(2)
    context_path_head, _ = os.path.split(context_path)

    # Handle reference outside document
    new_ref_context_path = os.path.join(context_path_head, ref_context)
    norm_new_ref_context_path = _norm_context(context=new_ref_context_path)
    return f"{context_hostname}{norm_new_ref_context_path}#{ref_schema}"


def _handle_match(match: typing.Match, *, context: str) -> str:
    """
    Map a match to the updated value.

    Args:
        match: The match to the regular expression for the reference.
        context: The context to use to update the reference.

    Returns:
        The updated reference.

    """
    ref = match.group(1)
    mapped_ref = _add_remote_context(context=context, ref=ref)
    return match.group(0).replace(ref, mapped_ref)


# Pattern used to look for any $ref after converting the schema to JSON
_REF_VALUE_PATTERN = re.compile(r'"\$ref": "(.*?)"')


def _map_remote_schema_ref(*, schema: types.Schema, context: str) -> types.Schema:
    """
    Update any $ref within the schema with the remote context.

    Serialize the schema, look for $ref and update value to include context.

    Args:
        schema: The schema to update.
        context: The context of the schema.

    Returns:
        The schema with any $ref mapped to include the context.

    """
    # Define context for mapping
    handle_match_context = functools.partial(_handle_match, context=context)

    str_schema = json.dumps(schema)
    mapped_str_schema = _REF_VALUE_PATTERN.sub(handle_match_context, str_schema)
    mapped_schema = json.loads(mapped_str_schema)
    return mapped_schema


class _RemoteSchemaStore:
    """Store remote schemas in memory to speed up use."""

    _schemas: typing.Dict[str, types.Schemas]
    spec_context: typing.Optional[str]

    def __init__(self) -> None:
        """Construct."""
        self._schemas = {}
        self.spec_context = None

    def reset(self):
        """Reset the state of the schema store."""
        self._schemas = {}
        self.spec_context = None

    def get_schemas(self, *, context: str) -> types.Schema:
        """
        Retrieve the schemas for a context.

        Raise MissingArgumentError if the context for the original OpenAPI specification
            has not been set.
        Raise SchemaNotFoundError if the context doesn't exist or is not a json nor yaml
            file.

        Args:
            context: The path, relative to the original OpenAPI specification, for the
                file containing the schemas.

        Returns:
            The schemas.

        """
        # Check whether the context is already loaded
        if context in self._schemas:
            return self._schemas[context]

        if self.spec_context is None:
            raise exceptions.MissingArgumentError(
                "Cannot find the file containing the remote reference, either "
                "initialize OpenAlchemy with init_json or init_yaml or pass the path "
                "to the OpenAPI specification to OpenAlchemy."
            )

        # Check for json, yaml or yml file extension
        _, extension = os.path.splitext(context)
        extension = extension.lower()
        if extension not in {".json", ".yaml", ".yml"}:
            raise exceptions.SchemaNotFoundError(
                "The remote context is not a JSON nor YAML file. The path is: "
                f"{context}"
            )

        # Get context manager with file
        try:
            if _URL_REF_PATTERN.search(context) is not None:
                file_cm = request.urlopen(context)
            else:
                spec_dir = os.path.dirname(self.spec_context)
                remote_spec_filename = os.path.join(spec_dir, context)
                file_cm = open(remote_spec_filename)
        except (FileNotFoundError, error.HTTPError) as exc:
            raise exceptions.SchemaNotFoundError(
                "The file with the remote reference was not found. The path is: "
                f"{context}"
            ) from exc

        # Calculate location of schemas
        with file_cm as in_file:
            if extension == ".json":
                try:
                    schemas = json.load(in_file)
                except json.JSONDecodeError as exc:
                    raise exceptions.SchemaNotFoundError(
                        "The remote reference file is not valid JSON. The path "
                        f"is: {context}"
                    ) from exc
            else:
                # Import as needed to make yaml optional
                import yaml  # pylint: disable=import-outside-toplevel

                try:
                    schemas = yaml.safe_load(in_file)
                except yaml.scanner.ScannerError as exc:
                    raise exceptions.SchemaNotFoundError(
                        "The remote reference file is not valid YAML. The path "
                        f"is: {context}"
                    ) from exc

        # Store for faster future retrieval
        self._schemas[context] = schemas
        return schemas


_remote_schema_store = _RemoteSchemaStore()  # pylint: disable=invalid-name


def set_context(*, path: str) -> None:
    """
    Set the context for the initial OpenAPI specification.

    Args:
        path: The path to the OpenAPI specification

    """
    _remote_schema_store.spec_context = path


def _retrieve_schema(*, schemas: types.Schemas, path: str) -> NameSchema:
    """
    Retrieve schema at a path from schemas.

    Raise SchemaNotFoundError if the schema is not found at the path.

    Args:
        schemas: All the schemas.
        path: The location to retrieve the schema from.

    Returns:
        The schema at the path from the schemas.

    """
    # Strip leading /
    if path.startswith("/"):
        path = path[1:]

    # Get the first directory/file as the head and the remaining path as the tail
    path_components = path.split("/", 1)

    try:
        # Base case, no tail
        if len(path_components) == 1:
            return path_components[0], schemas[path_components[0]]
        # Recursive case, call again with path tail
        return _retrieve_schema(
            schemas=schemas[path_components[0]], path=path_components[1]
        )
    except KeyError as exc:
        raise exceptions.SchemaNotFoundError(
            f"The schema was not found in the remote schemas. Path subsection: {path}"
        ) from exc


def get_remote_ref(*, ref: str) -> NameSchema:
    """
    Retrieve remote schema based on reference.

    Args:
        ref: The reference to the remote schema.

    Returns:
        The remote schema.

    """
    context, path = _separate_context_path(ref=ref)
    context = _norm_context(context=context)
    schemas = _remote_schema_store.get_schemas(context=context)
    name, schema = _retrieve_schema(schemas=schemas, path=path)
    schema = _map_remote_schema_ref(schema=schema, context=context)
    return name, schema



"""Types for validation."""

import typing

from ... import types


class Result(typing.NamedTuple):
    """Result of checking a schema."""

    # Whether the schema is valid
    valid: bool
    # If not valid, the reason why it isn't
    reason: typing.Optional[str]


OptResult = typing.Optional[Result]


class _TResultBase(types.TypedDict, total=False):
    """TResult base class for keys that might not be present."""

    reason: str


class TResult(_TResultBase, total=True):
    """Record the result of validation."""

    valid: bool


def t_result_from_result(result: Result) -> TResult:
    """Construct a result dictionary from a result."""
    return_dict: "TResult" = {"valid": result.valid}
    if result.reason is not None:
        return_dict["reason"] = result.reason

    return return_dict


class TProperty(types.TypedDict, total=True):
    """Record validation result of a property."""

    result: TResult


TProperties = typing.Dict[str, TProperty]


class _TModelBase(types.TypedDict, total=False):
    """TModel base class for keys that might not be present."""

    properties: TProperties


class TModel(_TModelBase, total=True):
    """Record validation result of a model."""

    result: TResult


TModels = typing.Dict[str, TModel]


class _TSpecBase(types.TypedDict, total=False):
    """TSpec base class for keys that might not be presents."""

    models: TModels


class TSpec(_TSpecBase, total=True):
    """Record validation result for a specification."""

    result: TResult



"""Functions for calculating the type."""

from open_alchemy import helpers as oa_helpers
from open_alchemy import types
from open_alchemy.schemas import artifacts as schemas_artifacts

_SIMPLE_TYPE_STRING_FORMAT_MAPPING = {
    "binary": "bytes",
    "date": "datetime.date",
    "date-time": "datetime.datetime",
}


_SIMPLE_TYPE_MAPPING = {
    "string": lambda format_: _SIMPLE_TYPE_STRING_FORMAT_MAPPING[format_]
    if format_ in _SIMPLE_TYPE_STRING_FORMAT_MAPPING
    else "str",
    "integer": lambda _: "int",
    "number": lambda _: "float",
    "boolean": lambda _: "bool",
}


def _model_simple_property(
    *, artifacts: schemas_artifacts.types.SimplePropertyArtifacts
) -> str:
    """Calculate the Python type of a simple property."""
    assert artifacts.open_api.type in oa_helpers.type_.SIMPLE_TYPES
    type_ = _SIMPLE_TYPE_MAPPING[artifacts.open_api.type](artifacts.open_api.format)
    optional = oa_helpers.calculate_nullable(
        nullable=artifacts.open_api.nullable,
        generated=artifacts.extension.autoincrement is True,
        required=artifacts.required,
        defaulted=artifacts.open_api.default is not None
        or artifacts.extension.server_default is not None,
    )

    if optional:
        return f"typing.Optional[{type_}]"
    return type_


_RELATIONSHIP_TYPE_MAPPING = {
    types.RelationshipType.MANY_TO_ONE: lambda parent: f'"T{parent}"',
    types.RelationshipType.ONE_TO_ONE: lambda parent: f'"T{parent}"',
    types.RelationshipType.ONE_TO_MANY: (
        lambda parent: f'typing.Sequence["T{parent}"]'
    ),
    types.RelationshipType.MANY_TO_MANY: (
        lambda parent: f'typing.Sequence["T{parent}"]'
    ),
}


def _model_relationship_property(
    *, artifacts: schemas_artifacts.types.TAnyRelationshipPropertyArtifacts
) -> str:
    """Calculate the Python type of a relationship property."""
    type_ = _RELATIONSHIP_TYPE_MAPPING[artifacts.sub_type](artifacts.parent)

    if (
        artifacts.sub_type == types.RelationshipType.ONE_TO_MANY
        or artifacts.sub_type == types.RelationshipType.MANY_TO_MANY
    ):
        return type_

    optional = oa_helpers.calculate_nullable(
        nullable=artifacts.nullable,
        generated=False,
        required=artifacts.required,
        defaulted=False,
    )

    if optional:
        return f"typing.Optional[{type_}]"
    return type_


def model(*, artifacts: schemas_artifacts.types.TAnyPropertyArtifacts) -> str:
    """
    Calculate the Python type of a property.

    Assume that the property is not a back reference.

    Args:
        artifacts: The artifacts for the property.

    Returns:
        The equivalent Python type.

    """
    assert artifacts.type != types.PropertyType.BACKREF

    if artifacts.type == types.PropertyType.SIMPLE:
        return _model_simple_property(artifacts=artifacts)

    if artifacts.type == types.PropertyType.RELATIONSHIP:
        return _model_relationship_property(artifacts=artifacts)

    return "typing.Any"


def typed_dict(*, artifacts: schemas_artifacts.types.TAnyPropertyArtifacts) -> str:
    """
    Calculate the Python type of a column for a TypedDict.

    Args:
        artifacts: The artifacts from the schema of the column.

    Returns:
        The equivalent Python type for the TypedDict.

    """
    model_type: str
    if artifacts.type != types.PropertyType.BACKREF:
        # Calculate type the same way as for the model
        model_type = model(artifacts=artifacts)
    else:
        inner_type = "typing.Dict[str, typing.Union[int, float, str, bool]]"
        if artifacts.sub_type == schemas_artifacts.types.BackrefSubType.OBJECT:
            model_type = f"typing.Optional[{inner_type}]"
        else:
            model_type = f"typing.Sequence[{inner_type}]"

    # No more checks if JSON
    if artifacts.type == types.PropertyType.JSON:
        return model_type

    # Modify the type in case of a relationship
    if artifacts.type == types.PropertyType.RELATIONSHIP:
        model_type = model_type.replace(
            f"T{artifacts.parent}", f"{artifacts.parent}Dict"
        )

    if artifacts.type == types.PropertyType.SIMPLE:
        # Revert back to str for binary, date and date-time
        if artifacts.open_api.format == "binary":
            model_type = model_type.replace("bytes", "str")
        if artifacts.open_api.format == "date":
            model_type = model_type.replace("datetime.date", "str")
        if artifacts.open_api.format == "date-time":
            model_type = model_type.replace("datetime.datetime", "str")

    return model_type


def arg_init(*, artifacts: schemas_artifacts.types.TAnyPropertyArtifacts) -> str:
    """
    Calculate the Python type of a column for the arguments of __init__.

    Assume that the property is not a back reference.

    Args:
        artifacts: The artifacts from the schema of the column.

    Returns:
        The equivalent Python type for the argument for the column.

    """
    assert artifacts.type != types.PropertyType.BACKREF

    model_type = model(artifacts=artifacts)

    # If has a default value, remove optional
    if artifacts.type == types.PropertyType.SIMPLE:
        if artifacts.open_api.default is not None:
            if not model_type.startswith("typing.Optional["):
                return model_type
            return model_type[16:-1]

    # Add optional if not required unless already optional
    if not artifacts.required and not model_type.startswith("typing.Optional["):
        return f"typing.Optional[{model_type}]"
    return model_type


def arg_from_dict(*, artifacts: schemas_artifacts.types.TAnyPropertyArtifacts) -> str:
    """
    Calculate the Python type of a column for the arguments of from_dict.

    Assume that the property is not a back reference.

    Args:
        artifacts: The artifacts from the schema of the column.

    Returns:
        The equivalent Python type for the argument for the column.

    """
    assert artifacts.type != types.PropertyType.BACKREF

    # Calculate type the same way as for the model
    init_type = arg_init(artifacts=artifacts)

    # No more checks if JSON
    if artifacts.type == types.PropertyType.JSON:
        return init_type

    # Modify the type in case of relationship
    if artifacts.type == types.PropertyType.RELATIONSHIP:
        init_type = init_type.replace(f"T{artifacts.parent}", f"{artifacts.parent}Dict")

    return init_type



"""Validate association tables."""

import typing

from ... import helpers as oa_helpers
from ... import types as oa_types
from .. import helpers
from . import helpers as validation_helpers
from . import types


class _TNameSchemaTablename(typing.NamedTuple):
    """Holds the name, schema and tablename of a schema."""

    name: str
    schema: oa_types.Schema
    tablename: str


def _get_defined_association_iterator(
    schemas: oa_types.Schemas, tablename_mapping: typing.Dict[str, typing.Any]
) -> typing.Iterable[_TNameSchemaTablename]:
    """
    Get an iterator with schemas that have a tablename as defined by a mapping.

    Assume that individual schemas are valid.

    Algorithm:
    1. iterate over constructable schemas
    2. include any with an x-tablename value that appears in the tablename mapping

    Args:
        schemas: All defined schemas.
        tablename_mapping: A mapping of tablename to some value. Keys are used to
            indicate to include a schema.

    Returns:
        An iterator with all schemas that have their x-tablename appear in the
        tablename mapping.

    """
    constructables = helpers.iterate.constructable(schemas=schemas)
    constructables_tablename = map(
        lambda args: _TNameSchemaTablename(
            name=args[0],
            schema=args[1],
            tablename=oa_helpers.peek.prefer_local(
                get_value=oa_helpers.peek.tablename, schema=args[1], schemas=schemas
            ),
        ),
        constructables,
    )
    return filter(
        lambda args: args.tablename in tablename_mapping,
        constructables_tablename,
    )


def _primary_key_property_items_iterator(
    schema: oa_types.Schema, schemas: oa_types.Schemas
) -> typing.Iterator[typing.Tuple[str, oa_types.Schema]]:
    """
    Get an iterable with only primary key properties.

    Args:
        schema: The schema to get primary key properties from.
        schemas: All defined schemas.

    Returns:
        An iterator with all primary key properties.

    """
    properties = helpers.iterate.properties_items(schema=schema, schemas=schemas)
    return filter(
        lambda args: oa_helpers.peek.prefer_local(
            get_value=oa_helpers.peek.primary_key,
            schema=args[1],
            schemas=schemas,
        ),
        properties,
    )


def _check_2_or_fewer_primary_key(
    name: str,
    schema: oa_types.Schema,
    association: helpers.association.TParentPropertySchema,
    schemas: oa_types.Schemas,
) -> types.Result:
    """
    Check that at most 2 primary key columns have been defined.

    Args:
        name: The name of the schema.
        schema: The schema to validate.
        association: Information about the association property.
        schemas: All defined schemas.

    Returns:
        Whether the schema is valid with a reason if it is not.

    """
    # Get first primary key property without foreign key
    primary_key_properties = _primary_key_property_items_iterator(schema, schemas)
    primary_key_count = sum(1 for _ in primary_key_properties)

    if primary_key_count > 2:
        return types.Result(
            valid=False,
            reason=(
                f'schema "{name}" defines more than 2 primary keys which is too many '
                "because it implements an association table for the many-to-many "
                f'relationship property "{association.property.name}" on the schema '
                f'"{association.parent.name}"'
            ),
        )

    return types.Result(valid=True, reason=None)


def _check_primary_key_no_foreign_key(
    name: str,
    schema: oa_types.Schema,
    association: helpers.association.TParentPropertySchema,
    schemas: oa_types.Schemas,
) -> types.Result:
    """
    Check that all primary keys define foreign keys.

    Args:
        name: The name of the schema.
        schema: The schema to validate.
        association: Information about the association property.
        schemas: All defined schemas.

    Returns:
        Whether the schema is valid with a reason if it is not.

    """
    # Get first primary key property without foreign key
    primary_key_properties = _primary_key_property_items_iterator(schema, schemas)
    not_foreign_key_primary_key_properties = filter(
        lambda args: oa_helpers.peek.foreign_key(schema=args[1], schemas=schemas)
        is None,
        primary_key_properties,
    )

    # Check for whether a primary key does not define a foreign key
    property_ = next(not_foreign_key_primary_key_properties, None)
    if property_ is not None:
        property_name, _ = property_
        return types.Result(
            valid=False,
            reason=(
                f'property "{property_name}" on schema "{name}" is a primary key and '
                "therefore must define a foreign key because it implements an "
                "association table for the many-to-many relationship property "
                f'"{association.property.name}" on the schema '
                f'"{association.parent.name}"'
            ),
        )

    return types.Result(valid=True, reason=None)


def _assert_str(item: typing.Any) -> str:
    """Check that item is a string."""
    assert isinstance(item, str)
    return item


def _check_duplicate_foreign_key(
    name: str,
    schema: oa_types.Schema,
    association: helpers.association.TParentPropertySchema,
    schemas: oa_types.Schemas,
) -> types.Result:
    """
    Check that all primary key foreign keys are unique.

    Assume that all primary keys define a foreign key.

    Args:
        name: The name of the schema.
        schema: The schema to validate.
        association: Information about the association property.
        schemas: All defined schemas.

    Returns:
        Whether the schema is valid with a reason if it is not.

    """
    # Get first primary key property without foreign key
    primary_key_properties = _primary_key_property_items_iterator(schema, schemas)
    foreign_keys = map(
        lambda args: (
            args[0],
            oa_helpers.peek.prefer_local(
                get_value=oa_helpers.peek.foreign_key, schema=args[1], schemas=schemas
            ),
        ),
        primary_key_properties,
    )
    # Check that items are a string
    foreign_keys = map(lambda args: (args[0], _assert_str(args[1])), foreign_keys)

    # Check for duplicate foreign keys
    seen_foreign_keys: typing.Dict[str, str] = {}
    for property_name, foreign_key in foreign_keys:
        if foreign_key not in seen_foreign_keys:
            seen_foreign_keys[foreign_key] = property_name
            continue

        seen_on_property_name = seen_foreign_keys[foreign_key]
        return types.Result(
            valid=False,
            reason=(
                f'duplicate foreign key "{foreign_key}" defined on a primary key '
                f'property "{property_name}" already defined on the property '
                f'"{seen_on_property_name}" on the schema "{name}" that implements an '
                "association table for the many-to-many relationship property "
                f'"{association.property.name}" on the schema '
                f'"{association.parent.name}"'
            ),
        )

    return types.Result(valid=True, reason=None)


def _check_properties_valid(
    name: str,
    schema: oa_types.Schema,
    association: helpers.association.TParentPropertySchema,
    schemas: oa_types.Schemas,
) -> types.Result:
    """
    Check that all primary key properties are valid.

    Assume that all primary keys define a foreign key.

    Check that:
    1. the primary keys have one of the expected foreign keys,
    2. that the primary key type matches and
    3. that the format, maxLength match the expectation, including whether they should
        be defined.

    Args:
        name: The name of the schema.
        schema: The schema to validate.
        association: Information about the association property.
        schemas: All defined schemas.

    Returns:
        Whether the schema is valid with a reason if it is not.

    """
    # Calculate the expected schema
    expected_schema = helpers.association.calculate_schema(
        property_schema=association.property.schema,
        parent_schema=association.parent.schema,
        schemas=schemas,
    ).schema
    # Creating a mapping of foreign key to expected property schema
    expected_foreign_key_properties = {
        property_["x-foreign-key"]: property_
        for property_ in expected_schema["properties"].values()
    }

    # Check primary keys
    primary_key_properties = _primary_key_property_items_iterator(schema, schemas)
    for property_name, property_schema in primary_key_properties:
        property_foreign_key = oa_helpers.peek.prefer_local(
            get_value=oa_helpers.peek.foreign_key,
            schema=property_schema,
            schemas=schemas,
        )

        # Check that the foreign key is expected
        if property_foreign_key not in expected_foreign_key_properties:
            expected_foreign_keys = ", ".join(
                map(lambda key: f'"{key}"', expected_foreign_key_properties.keys())
            )
            return types.Result(
                valid=False,
                reason=(
                    f'unexpected foreign key "{property_foreign_key}" defined on a '
                    f'primary key property "{property_name}" on the schema "{name}" '
                    "that implements an association table for the many-to-many "
                    f'relationship property "{association.property.name}" on the '
                    f'schema "{association.parent.name}", expected one of '
                    f"{expected_foreign_keys}"
                ),
            )

        # Check that the property schema is as expected
        expected_schema = expected_foreign_key_properties[property_foreign_key]
        checks = (
            ("type", oa_helpers.peek.type_),
            ("format", oa_helpers.peek.format_),
            ("maxLength", oa_helpers.peek.max_length),
        )
        for key, func in checks:
            # Check that values match
            result = validation_helpers.value.check_matches(
                func=func,
                reference_schema=expected_schema,
                check_schema=property_schema,
                schemas=schemas,
            )
            if result is None:
                continue

            return types.Result(
                valid=False,
                reason=(
                    f"unexpected {key} ({result}) defined on a "
                    f'primary key property "{property_name}" on the schema "{name}" '
                    "that implements an association table for the many-to-many "
                    f'relationship property "{association.property.name}" on the '
                    f'schema "{association.parent.name}"'
                ),
            )

    return types.Result(valid=True, reason=None)


def _validate_schema(
    name: str,
    schema: oa_types.Schema,
    association: helpers.association.TParentPropertySchema,
    schemas: oa_types.Schemas,
) -> types.Result:
    """
    Validate the schema for an association against the expected schema.

    Assume that schemas are individually valid.

    A know limitation is that the following invalid schema would pass:
    Parent:
        x-tablename: parent
    Child1:
        allOf:
            -   $ref: "#/components/schemas/Parent"
            -   properties:
                    x-inherits: true
                    prop_1:
                        primary_key: true
                        x-foreign-key: foreign.key
    Child2:
        allOf:
            -   $ref: "#/components/schemas/Parent"
            -   properties:
                    x-inherits: true
                    prop_2:
                        primary_key: true
                        x-foreign-key: foreign.key
    The problem is that both Child1 and Child2 define the same foreign key but are
    different schemas. Schemas are considered in isolation.

    Args:
        name: The name of the schema.
        schema: The schema to validate.
        association: The information about the many-to-many property.
        schemas: All defined schemas.

    """
    checks = [
        _check_2_or_fewer_primary_key,
        _check_primary_key_no_foreign_key,
        _check_duplicate_foreign_key,
        _check_properties_valid,
    ]
    for validation_check in checks:
        result = validation_check(name, schema, association, schemas)
        if not result.valid:
            return result

    return types.Result(valid=True, reason=None)


def check(*, schemas: oa_types.Schemas) -> types.Result:
    """
    Check that any schemas that implement association tables are valid.

    Assume that schemas are otherwise valid.

    The algorithm is:
    1. get a mapping of x-secondary to property information,
    2. iterate over schemas and retrieve all that have define an association and
    3. validate those schemas.

    Args:
        schemas: The schemas to check.

    Returns:
        Whether the schemas that define association tables are valid.

    """
    # Get mapping
    secondary_parent_property_schema_mapping = (
        helpers.association.get_secondary_parent_property_schema_mapping(
            schemas=schemas
        )
    )

    # Get association schemas
    association_schemas = _get_defined_association_iterator(
        schemas, secondary_parent_property_schema_mapping
    )

    # Validate schemas
    for association_schema in association_schemas:
        association_info = secondary_parent_property_schema_mapping[
            association_schema.tablename
        ]
        result = _validate_schema(
            association_schema.name,
            association_schema.schema,
            association_info,
            schemas,
        )
        if not result.valid:
            return result

    return types.Result(valid=True, reason=None)



"""Define the CLI module."""
import argparse
import logging
import pathlib

from open_alchemy import PackageFormat
from open_alchemy import build_json
from open_alchemy import build_yaml
from open_alchemy import exceptions
from open_alchemy import init_json
from open_alchemy import init_yaml

# Configure the logger.
logging.basicConfig(format="%(message)s", level=logging.INFO)

# Define constants.
VALID_EXTENSIONS = [".json", ".yml", ".yaml"]


def main() -> None:
    """Define the CLI entrypoint."""
    args = build_application_parser()
    try:
        args.func(args)
    except exceptions.CLIError as exc:
        logging.error("Cannot perform the operation: %s.", exc)


def validate_specfile(specfile: pathlib.Path) -> None:
    """
    Ensure the specfile is valid.

    The specification file must:
        * exist
        * have a valid extension (.json, .yml, or .yaml)

    Args:
        specfile: Path object representing the specification file.
    """
    # Ensure the file exists.
    if not specfile.exists():
        raise exceptions.CLIError(f"the specfile '{specfile}' cannot be found")

    # Ensure the extension is valid.
    ext = specfile.suffix.lower()
    if ext not in VALID_EXTENSIONS:
        raise exceptions.CLIError(
            f"specification format not supported: '{specfile}' "
            f"(valid extensions are {', '.join(VALID_EXTENSIONS)}, "
            "case insensitive)"
        )


def build_application_parser() -> argparse.Namespace:
    """Build the main parser and subparsers."""
    # Define the top level parser.
    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers(title="subcommands", help="subcommand help")

    # Define the parser for the "build" subcommand.
    build_parser = subparsers.add_parser(
        "build",
        description="Build a package with the SQLAlchemy models.",
        help="build a package",
    )
    build_parser.add_argument(
        "specfile", type=str, help="specify the specification file"
    )
    build_parser.add_argument("name", type=str, help="specify the package name")
    build_parser.add_argument("output", type=str, help="specify the output directory")
    build_parser.add_argument(
        "-f",
        "--format",
        type=str.lower,
        choices=["sdist", "wheel"],
        help="limit the format to either sdist or wheel, defaults to both",
    )
    build_parser.set_defaults(func=build)

    # Define the parser for the "generate" subcommand.
    generate_parser = subparsers.add_parser(
        "generate",
        description="Generate the SQLAlchemy models.",
        help="regenerate models",
    )
    generate_parser.add_argument(
        "specfile", type=str, help="specify the specification file"
    )
    generate_parser.add_argument("output", type=str, help="specify the output file")
    generate_parser.set_defaults(func=generate)

    # Return the parsed arguments for a particular command.
    return parser.parse_args()


def build(args: argparse.Namespace) -> None:
    """
    Define the build subcommand.

    Args:
        args: CLI arguments from the parser.
    """
    # Check the specfile.
    specfile = pathlib.Path(args.specfile)
    validate_specfile(specfile)

    # Select the builder method.
    builders = dict(zip(VALID_EXTENSIONS, [build_json, build_yaml, build_yaml]))

    # Choose the distributable format.
    fmt = {
        None: PackageFormat.SDIST | PackageFormat.WHEEL,
        "sdist": PackageFormat.SDIST,
        "wheel": PackageFormat.WHEEL,
    }

    # Build the package.
    builder = builders.get(specfile.suffix.lower())
    builder(args.specfile, args.name, args.output, fmt.get(args.format))  # type: ignore


def generate(args):
    """
    Define the generate subcommand.

    Args:
        args: CLI arguments from the parser.
    """
    # Check the specfile.
    specfile = pathlib.Path(args.specfile)
    validate_specfile(specfile)

    # Select the generator method.
    generators = dict(zip(VALID_EXTENSIONS, [init_json, init_yaml, init_yaml]))

    # Regenerate the models.
    generator = generators.get(specfile.suffix.lower())
    generator(args.specfile, models_filename=args.output)



"""Create table args such as Uniques and Index."""

import functools
import itertools
import json
import os
import typing

from sqlalchemy import schema

from open_alchemy import exceptions
from open_alchemy import facades
from open_alchemy import types

_DIRECTORY = os.path.dirname(__file__)
_PATHS = ("..", "helpers", "ext_prop")
_COMMON_SCHEMAS_FILE = os.path.join(_DIRECTORY, *_PATHS, "common-schemas.json")
(
    _resolver,  # pylint: disable=invalid-name
    (_COMMON_SCHEMAS,),
) = facades.jsonschema.resolver(_COMMON_SCHEMAS_FILE)


def _spec_to_schema_name(
    *,
    spec: typing.Union[types.AnyUnique, types.AnyIndex],
    schema_names: typing.Optional[typing.List[str]] = None
) -> str:
    """
    Convert a specification to the name of the matched schema.

    Use the schema names defined in common-schemas.json to find the first matching
    schema.

    Args:
        spec: The specification to convert.
        schema_names: The names of the schemas to check.

    Returns:
        The name of the specification.

    """
    if schema_names is None:
        schema_names = list(_COMMON_SCHEMAS.keys())

    for name in schema_names:
        try:
            facades.jsonschema.validate(
                instance=spec, schema=_COMMON_SCHEMAS[name], resolver=_resolver
            )
            return name
        except facades.jsonschema.ValidationError:
            continue
    raise exceptions.SchemaNotFoundError("Specification did not match any schemas.")


def _handle_column_list(spec, property_name):
    """Convert ColumnList to a dictionary."""
    return {property_name: spec}


# Handling for column lists for unique constraints and index
_uq_handle_column_list = functools.partial(  # pylint: disable=invalid-name
    _handle_column_list, property_name="columns"
)
_ix_handle_column_list = functools.partial(  # pylint: disable=invalid-name
    _handle_column_list, property_name="expressions"
)

# Schema names for unique constraints and index
_SCHEMAS_FILE = os.path.join(_DIRECTORY, *_PATHS, "extension-schemas.json")
with open(_SCHEMAS_FILE) as in_file:
    _SCHEMAS = json.load(in_file)
_UNIQUE_SCHEMA_NAMES: typing.List[str] = list(
    map(
        lambda schema: schema["$ref"].split("/")[-1],
        _COMMON_SCHEMAS["CompositeUnique"]["oneOf"],
    )
)
_INDEX_SCHEMA_NAMES: typing.List[str] = list(
    map(
        lambda schema: schema["$ref"].split("/")[-1],
        _COMMON_SCHEMAS["CompositeIndex"]["oneOf"],
    )
)

# Unique and index name to conversion function
_UNIQUE_MAPPING: typing.Dict[str, typing.Callable[..., types.UniqueList]] = {
    "ColumnList": lambda spec: [_uq_handle_column_list(spec=spec)],
    "ColumnListList": lambda spec: list(map(_uq_handle_column_list, spec)),
    "Unique": lambda spec: [spec],
    "UniqueList": lambda spec: spec,
}
_INDEX_MAPPING: typing.Dict[str, typing.Callable[..., types.IndexList]] = {
    "ColumnList": lambda spec: [_ix_handle_column_list(spec=spec)],
    "ColumnListList": lambda spec: list(map(_ix_handle_column_list, spec)),
    "Index": lambda spec: [spec],
    "IndexList": lambda spec: spec,
}


def map_unique(*, spec: types.AnyUnique) -> types.UniqueList:
    """
    Convert any unique constraint to UniqueList.

    Args:
        spec: The specification to convert.

    Returns:
        The UniqueList.

    """
    name = _spec_to_schema_name(spec=spec, schema_names=_UNIQUE_SCHEMA_NAMES)
    return _UNIQUE_MAPPING[name](spec)


def iter_unique_columns(*, spec: types.AnyUnique) -> typing.Iterator[str]:
    """
    Iterate over all columns in the composite unique constraint.

    Args:
        spec: The specification to iterator over.

    Returns:
        An iterator with all columns of the unique constraint.

    """
    mapped_specs = map_unique(spec=spec)
    column_lists = map(lambda unique_spec: unique_spec["columns"], mapped_specs)
    return itertools.chain(*column_lists)


def map_index(*, spec: types.AnyIndex) -> types.IndexList:
    """
    Convert any composite index to IndexList.

    Args:
        spec: The specification to convert.

    Returns:
        The IndexList.

    """
    name = _spec_to_schema_name(spec=spec, schema_names=_INDEX_SCHEMA_NAMES)
    return _INDEX_MAPPING[name](spec)


def iter_index_expressions(*, spec: types.AnyIndex) -> typing.Iterator[str]:
    """
    Iterate over all expressions in the composite index constraint.

    Args:
        spec: The specification to iterator over.

    Returns:
        An iterator with all expressions of the index constraint.

    """
    mapped_specs = map_index(spec=spec)
    column_lists = map(lambda index_spec: index_spec["expressions"], mapped_specs)
    return itertools.chain(*column_lists)


def _construct_unique(spec: types.Unique) -> schema.UniqueConstraint:
    """
    Construct unique constraints.

    Args:
        spec: The definitions for the unique constraints.

    Returns:
        The unique constraints.

    """
    name = spec.get("name")
    columns = spec["columns"]

    return schema.UniqueConstraint(*columns, name=name)


def _construct_index(spec: types.Index) -> schema.Index:
    """
    Construct composite index.

    Args:
        spec: The definitions for the composite index.

    Returns:
        The composite index.

    """
    # There is a bug in the sqlalchemy-stubs where name is not Optional for Index
    name: str = spec.get("name")  # type: ignore
    unique = spec.get("unique")
    expressions = spec["expressions"]

    if unique is not None:
        return schema.Index(name, *expressions, unique=unique)
    return schema.Index(name, *expressions)


def unique_factory(
    *, spec: types.AnyUnique
) -> typing.Iterator[schema.UniqueConstraint]:
    """
    Generate unique constraints from specification.

    Args:
        spec: The specification to use.

    Returns:
        The unique constraints.

    """
    mapped_spec = map_unique(spec=spec)
    return map(_construct_unique, mapped_spec)


def index_factory(*, spec: types.AnyIndex) -> typing.Iterator[schema.Index]:
    """
    Generate composite indexes from specification.

    Args:
        spec: The specification to use.

    Returns:
        The composite indexes.

    """
    mapped_spec = map_index(spec=spec)
    return map(_construct_index, mapped_spec)



"""Validation for properties."""

from .... import exceptions
from .... import helpers as oa_helpers
from .... import types as oa_types
from .. import types
from . import backref
from . import json
from . import relationship
from . import simple

Type = oa_types.PropertyType


def check_type(schemas: oa_types.Schemas, schema: oa_types.Schema) -> types.Result:
    """
    Check whether the type of a property can be calculated.

    Algorithm:
    1. check that the type is present and valid
    2. check that it is one of integer, number, string, boolean, object or array,
    3. check that x-json, if defined, is valid and
    3. check that readOnly, if defined, is valid.

    Args:
        schemas: All defined schemas used to resolve any $ref.
        schema: The property schema to check.

    Returns:
        Whether thetype of the property can be determined.

    """
    try:
        type_ = oa_helpers.peek.type_(schema=schema, schemas=schemas)
        if type_ not in oa_helpers.type_.TYPES:
            return types.Result(False, f"{type_} is not a supported type")
        oa_helpers.peek.json(schema=schema, schemas=schemas)
        oa_helpers.peek.read_only(schema=schema, schemas=schemas)
        oa_helpers.peek.write_only(schema=schema, schemas=schemas)

    except (exceptions.MalformedSchemaError, exceptions.TypeMissingError) as exc:
        return types.Result(False, f"malformed schema :: {exc}")
    except exceptions.SchemaNotFoundError as exc:
        return types.Result(False, f"reference :: {exc}")

    return types.Result(True, None)


def check(
    schemas: oa_types.Schemas,
    parent_schema: oa_types.Schema,
    property_name: str,
    property_schema: oa_types.Schema,
) -> types.Result:
    """
    Check the schema for a property.

    Args:
        schemas: All defined schemas used to resolve any $ref.
        parent_schema: The schema the property is embedded in.
        property_name: The name of the property.
        property_schema: The schema to check.

    Returns:
        Whether the property is valid.

    """
    type_result = check_type(schema=property_schema, schemas=schemas)
    if not type_result.valid:
        return type_result

    type_ = oa_helpers.property_.calculate_type(schema=property_schema, schemas=schemas)

    if type_ == Type.BACKREF:
        return backref.check(schema=property_schema, schemas=schemas)
    if type_ == Type.SIMPLE:
        return simple.check(schema=property_schema, schemas=schemas)
    if type_ == Type.JSON:
        return json.check(schema=property_schema, schemas=schemas)
    return relationship.check(
        property_name=property_name,
        property_schema=property_schema,
        parent_schema=parent_schema,
        schemas=schemas,
    )



"""Functions to convert to dictionary."""

import typing

from ... import exceptions
from ... import helpers
from ... import types as oa_types
from .. import types
from . import array
from . import object_
from . import simple


def convert(*, schema: oa_types.Schema, value: typing.Any) -> types.TAnyDict:
    """
    Convert value for a schema to a dictionary.

    Args:
        value: The value to convert.
        schema: The schema of the value.

    Returns:
        The converted value.

    """
    json = helpers.peek.json(schema=schema, schemas={})
    if json:
        return value
    type_ = helpers.peek.type_(schema=schema, schemas={})
    if type_ == "object":
        return object_.convert(value, schema=schema)
    if type_ == "array":
        return array.convert(value, schema=schema)
    if type_ in helpers.type_.SIMPLE_TYPES:
        return simple.convert(value, schema=schema)
    raise exceptions.FeatureNotImplementedError(f"Type {type_} is not supported.")


def return_none(*, schema: oa_types.Schema, property_name: str) -> bool:
    """
    Check whether a null value for a property should be returned.

    Assume the schema has properties and that it has a schema for the property.
    Assume that any $ref and allOf has already been resolved.

    The rules are:
    1. if the property is required, return it,
    2. if the property is nullable, return it and
    3. else, don't return it.

    Args:
        schema: The schema for the model.
        property_name: The name of the property to check for.

    Returns:
        Whether the none value should be returned for the property.

    """
    # Retrieve input
    required_array = schema.get("required", None)
    property_schema = schema["properties"][property_name]

    # Check for required and nullable
    if required_array is not None and property_name in set(required_array):
        return True
    nullable_value = helpers.peek.nullable(schema=property_schema, schemas={})
    if nullable_value is True:
        return True
    return False



"""Validate the specification."""

import typing

from . import schemas_validation
from . import types


def check(*, spec: typing.Any) -> types.Result:
    """
    Check whether a specification is valid.

    The algorithm is:
    1. check that it is a dictionary,
    2. check that components is defined and that it is a dictionary and
    3. check that schemas is defined and that it is valid.

    Args:
        spec: The specification to check.

    Returns:
        Whether the specification is valid.

    """
    if not isinstance(spec, dict):
        return types.Result(False, "specification must be a dictionary")

    # Check components
    components = spec.get("components")
    if components is None:
        return types.Result(False, "specification must define components")
    if not isinstance(components, dict):
        return types.Result(False, "components value must be a dictionary")

    # Check schemas
    schemas = components.get("schemas")
    if schemas is None:
        return types.Result(False, "specification must define schemas")
    return schemas_validation.check(schemas=schemas)



"""Define validation rules for relationship."""


from ..... import types as oa_types
from ... import types
from . import full
from . import property_


def check(
    schemas: oa_types.Schemas,
    parent_schema: oa_types.Schema,
    property_name: str,
    property_schema: oa_types.Schema,
) -> types.Result:
    """
    Check the schema of a relationship.

    Args:
        schemas: All the schemas used to resolve any $ref.
        parent_schema: The schema with the embedded relationship property.
        property_name: The name of the relationship property.
        property_schema: The schema of the relationship property.

    Returns:
        Whether the schema is valid and the reason if it is not.

    """
    # Validate the property schema
    property_result = property_.check(schemas, property_schema)
    if not property_result.valid:
        return property_result

    # Validate the full schema
    return full.check(schemas, parent_schema, property_name, property_schema)



"""Facade for jsonschema."""

import functools
import json
import typing

import jsonschema

# Re mapping values
ValidationError = jsonschema.ValidationError
validate = jsonschema.validate  # pylint: disable=invalid-name


def _filename_to_dict(filename: str) -> typing.Dict:
    """
    Map filename for a JSON file to the de-serialized dictionary.

    Args:
        filename: The name of the JSON file.

    Returns:
        The de-serialized contents of the file as a dictionary.

    """
    with open(filename) as in_file:
        json_dict = json.loads(in_file.read())
    return json_dict


def resolver(
    *filenames: str,
) -> typing.Tuple[
    jsonschema.RefResolver, typing.Tuple[typing.Dict[str, typing.Any], ...]
]:
    """
    Create resolver for references to schemas in another file.

    Args:
        filenames: The names for the files to add to the resolver.

    Returns:
        The resolver and the underlying schemas as a dictionary.

    """
    schema_dicts = tuple(map(_filename_to_dict, filenames))
    initial: typing.Dict[str, typing.Any] = {}
    merged_schema = functools.reduce(lambda x, y: {**x, **y}, schema_dicts, initial)
    return jsonschema.RefResolver.from_schema(merged_schema), schema_dicts



"""Helpers to determine the type of a schema."""

from .. import types
from . import peek
from . import type_


def calculate_type(
    *, schemas: types.Schemas, schema: types.Schema
) -> types.PropertyType:
    """
    Calculate the type of the property.

    Assume the property has a valid type.

    The rules are:
    1. if x-json is True it is JSON,
    2. if the type is integer, number, string or boolean it is SIMPLE,
    3. if readOnly is true it is BACKREF and
    4. otherwise it is RELATIONSHIP.

    Args:
        schemas: All defined schemas used to resolve any $ref.
        schema: The schema to calculate the type for.

    Returns:
        The type of the property.

    """
    json_value = peek.json(schema=schema, schemas=schemas)
    if json_value is True:
        return types.PropertyType.JSON

    property_type = peek.type_(schema=schema, schemas=schemas)
    if property_type in type_.SIMPLE_TYPES:
        return types.PropertyType.SIMPLE

    read_only_value = peek.read_only(schema=schema, schemas=schemas)
    if read_only_value is True:
        return types.PropertyType.BACKREF

    return types.PropertyType.RELATIONSHIP



"""Functions for interacting with the OpenAlchemy models."""

import typing

import open_alchemy
from open_alchemy import types

from ..utility_base import TOptUtilityBase
from ..utility_base import TUtilityBase


def get_base() -> typing.Any:
    """
    Get the models.Base used as the declarative base for models.

    Returns:
        The models.Base.

    """
    # pylint: disable=no-member
    return open_alchemy.models.Base  # type: ignore


def get_model(*, name: str) -> TOptUtilityBase:
    """
    Get a model by name from models.

    Args:
        name: The name of the model.

    Returns:
        The model with the name.

    """
    return getattr(open_alchemy.models, name, None)


def get_model_schema(*, name: str) -> typing.Optional[types.Schema]:
    """
    Get the schema of a model by name from models.

    Args:
        name: The name of the model.

    Returns:
        The schema of the model with the name.

    """
    model = get_model(name=name)
    if model is None:
        return None
    return model._schema  # pylint: disable=protected-access


def set_model(*, name: str, model: TUtilityBase) -> None:
    """
    Set model by name on models.

    Args:
        model: The model to set.
        name: The name of the model.

    """
    setattr(open_alchemy.models, name, model)



"""Generate the models file."""
# pylint: disable=useless-import-alias

import dataclasses
import typing

from open_alchemy import facades
from open_alchemy import schemas
from open_alchemy import types as oa_types

from . import artifacts as _artifacts
from . import model as _model
from . import models as _models
from . import types as types


def generate(*, artifacts: schemas.artifacts.types.ModelsModelArtifacts) -> str:
    """
    Generate the models file from schema artifacts.

    Args:
        artifacts: The artifacts from the schemas.

    Returns:
        The models file.

    """
    model_sources = list(
        map(
            lambda args: _model.generate(artifacts=args[1], name=args[0]),
            artifacts.items(),
        )
    )
    raw_source = _models.generate(models=model_sources)
    return facades.code_formatter.apply(source=raw_source)



"""Helpers for properties."""

from .... import types as oa_types
from ... import helpers
from .. import types


def check_properties_values(
    *, schema: oa_types.Schema, schemas: oa_types.Schemas
) -> types.OptResult:
    """
    Check the values of the properties.

    Args:
        schema: The schema to check.
        schemas: All defined schemas used to resolve any $ref.

    Returns:
        The result if the properties values are not valid with a reason or None.

    """
    properties_values = helpers.iterate.properties_values(
        schema=schema, schemas=schemas
    )
    not_dict_value = next(
        filter(lambda arg: not isinstance(arg, dict), properties_values), None
    )
    if not_dict_value is not None:
        return types.Result(False, "value of properties must be a dictionary")
    return None


def check_properties_items(
    *, schema: oa_types.Schema, schemas: oa_types.Schemas
) -> types.OptResult:
    """
    Check the items of the properties.

    Args:
        schema: The schema to check.
        schemas: All defined schemas used to resolve any $ref.

    Returns:
        The result if the properties items are not valid with a reason or None.

    """
    # Check names are string and values are dictionaries
    properties_items = helpers.iterate.properties_items(schema=schema, schemas=schemas)

    def check_property(args) -> types.OptResult:
        """Check key and value of property."""
        key, value = args
        if not isinstance(key, str):
            return types.Result(False, f"property names must be strings, {key} is not")
        if not isinstance(value, dict):
            return types.Result(False, f"{key} :: property values must be dictionaries")
        return None

    # Check for any results
    properties_items_results = map(check_property, properties_items)
    first_result = next(filter(None, properties_items_results), None)
    if first_result is not None:
        return first_result

    return None



"""Construct relationship properties."""

import typing

from sqlalchemy import orm

from open_alchemy import types as oa_types

from . import types


def construct(
    *, artifacts: oa_types.TAnyRelationshipPropertyArtifacts
) -> types.Relationship:
    """
    Construct relationship from artifacts.

    Args:
        artifacts: The artifacts of the relationship.

    Returns:
        The SQLAlchemy relationship.

    """
    # Construct back reference
    backref = None
    if artifacts.backref_property is not None:
        # Calculate uselist
        uselist: typing.Optional[bool] = None
        if artifacts.sub_type == oa_types.RelationshipType.ONE_TO_ONE:
            uselist = False

        backref = orm.backref(
            artifacts.backref_property,
            uselist=uselist,
        )

    # Calculate secondary
    secondary: typing.Optional[str] = None
    if artifacts.sub_type == oa_types.RelationshipType.MANY_TO_MANY:
        secondary = artifacts.secondary

    # Construct kwargs
    kwargs: typing.Dict[str, typing.Any] = {}
    if artifacts.kwargs is not None:
        kwargs = artifacts.kwargs

    # Construct relationship
    return orm.relationship(
        artifacts.parent, backref=backref, secondary=secondary, **kwargs
    )



"""Calculate whether a column is nullable."""

import typing


def calculate_nullable(
    *,
    nullable: typing.Optional[bool],
    generated: bool,
    defaulted: bool,
    required: typing.Optional[bool],
) -> bool:
    """
    Calculate the value of the nullable field.

    The following is the truth table for the nullable property.
    +------------+-----------+---------+-----------------+-------------------+
    | required   | generated | default | schema nullable | returned nullable |
    +============+===========+=========+=================+===================+
    | None/False | None      | None    | not given       | True              |
    +------------+-----------+---------+-----------------+-------------------+
    | None/False | None      | given   | not given       | False             |
    +------------+-----------+---------+-----------------+-------------------+
    | None/False | False     | X       | not given       | True              |
    +------------+-----------+---------+-----------------+-------------------+
    | None/False | True      | X       | not given       | False             |
    +------------+-----------+---------+-----------------+-------------------+
    | True       | X         | X       | not given       | False             |
    +------------+-----------+---------+-----------------+-------------------+
    | X          | X         | X       | False           | False             |
    +------------+-----------+---------+-----------------+-------------------+
    | X          | X         | X       | True            | True              |
    +------------+-----------+---------+-----------------+-------------------+

    To summarize, if nullable is the schema the value for it is used. Otherwise True
    is returned unless required is True, generated is True or there is a default value.

    Args:
        nullable: Whether the property is nullable.
        generated: Whether the column gets generated in some form (eg. auto increment,
            default value).
        required: Whether the property is required.

    Returns:
        The nullable value for the column.

    """
    if nullable is None:
        if required:
            return False
        if generated:
            return False
        if defaulted:
            return False
        return True
    if nullable:
        return True
    return False



"""Construct column for json property."""

from .. import facades
from .. import types as oa_types
from . import types


def handle(*, artifacts: oa_types.JsonPropertyArtifacts) -> types.TColumn:
    """
    Handle a json property.

    Args:
        artifacts: The artifacts of the json property.

    Returns:
        The constructed column.

    """
    return facades.sqlalchemy.json.construct(artifacts=artifacts)



"""SQLAlchemy types."""

import sqlalchemy
from sqlalchemy import orm

from ... import types as oa_types

# Remapping SQLAlchemy classes
Column = sqlalchemy.Column
Type = sqlalchemy.sql.type_api.TypeEngine
ForeignKey = sqlalchemy.ForeignKey
Integer = sqlalchemy.Integer
BigInteger = sqlalchemy.BigInteger
Number = sqlalchemy.Float
String = sqlalchemy.String
Binary = sqlalchemy.LargeBinary
Date = sqlalchemy.Date
DateTime = sqlalchemy.DateTime
Boolean = sqlalchemy.Boolean
JSON = sqlalchemy.JSON
Relationship = orm.RelationshipProperty


class TOptColumnArgs(oa_types.TypedDict, total=False):
    """Keyword arguments for Column."""

    primary_key: bool
    autoincrement: bool
    index: bool
    unique: bool



"""Types for column factory."""

import typing

from open_alchemy import facades

TColumn = facades.sqlalchemy.types.Column
TRelationship = facades.sqlalchemy.types.Relationship
TReturnValue = typing.Optional[typing.Union[TColumn, TRelationship]]
